/*==============================================================================


   Purpose:    Performance Counter Library (PCL)


   Author:     Heinz Ziegler, Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   22.07.1999


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 1999, Forschungszentrum Juelich GmbH, Federal Republic of
   Germany. All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.
  
==============================================================================*/


#include "pcl.h"
#include "determine_mhz.h"
#include "processor_number.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#if defined(JAVA_INTERFACE)
/* the include file PCL.h gets automatically generated by a call to javah */
#include "PCL.h"
#include <jni.h>
#endif


/*                                                                            */
/*============================================================================*/
/* Global definitions. Read carefully the description and decide if you want  */
/* to change any of them.                                                     */
/* Defaults are set in such way that you always get correct results on all    */
/* systems. I you are not sure, don't touch the definitions.                  */


/* Our performance counter result values are 64 bit integers. On some systems,
   hardware counters are only 32 bit counters (see below). Therefore, on such
   systems we activate a timer which interrupts and we gather temporary
   results. For such systems, COUNTER_SIZE_32 must be defined (below).
   The timer which gets activated interrupts every PCL_INTERRUPT_INTERVALL
   milliseconds to gather performance data. If you know the MHZ rate or cycle
   time of a processor, this could be more accurately set for every processor.
   Change the following definition if your processor can generate more than 2^32
   events per second (e.g. if you have a processor running a 5 Gigaherz :-) */
#define PCL_INTERRUPT_INTERVALL   1000     /* interrupt intervall (in msec) */

/* Normally, we decrement a real time counter to interrupt after
   PCL_INTERRUPT_INTERVALL msec. Doing so, the signal SIGALRM is used.
   If for some special reason you don't like that and prefer virtual
   processor time delivering the signal SIGVTALRM change the following
   line to "#if 0".  But please be aware that for an elapsed cycle counter
   counting is going on even your process is not scheduled by the OS.
  */
#if 1
#define ITIMER_MODE   ITIMER_REAL     /* decrement in real time */
#define ITIMER_SIGNAL SIGALRM         /* SIGALRM is delivered */
#else
#define ITIMER_MODE   ITIMER_VIRTUAL  /* decrement virtual process time */
#define ITIMER_SIGNAL SIGVTALRM       /* SIGVTALRM is delivered */
#endif


/* Undefine this macro if you have 32-bit counters but your measurements
   don't exceed 32-bit counters. If you undefine this macro, timers don't
   need to be set up and you can save a lot of overhead in PCLstart/PCLstop.
*/
#define COUNTER32_ARE_NOT_ENOUGH


/* Undefine this macro if you don't want to check consistency of function
   arguments, i.e. you are sure always to deliver correct arguments to the
   library interface functions.
*/
#define CHECK_ARGUMENTS


/* test if bit "bit" is set */
#define IS_USED(mask,bit)     (mask & (1u << bit))
/* set bit "bit"*/
#define DEFINE_USE(mask, bit) (mask |= (1u << bit))
/* reset bit "bit"*/
#define RESET_USE(mask, bit)  (mask &= ~(1u << bit))


/*                                                                            */
/*============================================================================*/
/* Include hardware/os dependend libraries, global variables and definitions  */

/* For every processor, PCL_MAX_COUNTER must be defined which gives the maximum
   number of hardware performance counters. If you summarize several processors
   families with a different number of performance counters under one category,
   take the maximum value.  Some systems have a special cycle counter for
   elapsed cycles. You should include this counter in the definition of
   PCL_MAX_COUNTER.
*/


/*----------------------------------------------------------------------------*/
/* settings for Alpha_21164-Unicos/mk-SGI/CRAY. */

#if defined(ALPHA_CRAY)

/* define the following macro if your T3E system might do a process migration */
/*#define CHECK_MIGRATION*/

#define PCL_MAX_COUNTER 4         /* 21164 has 3 counters and a cycle counter */

/* although the hardware performance counters are 64 bit counters, the
   processor cycle counter has only 32 bits.
*/
#define COUNTER_SIZE_32
#include <mpp/globals.h>
#include <unistd.h>
#include "machines/ALPHA_CRAY/perfctr.h"

/* shorts do best */
typedef unsigned short bit_mask_t;

/* initialized? */
static int initialized = 0;
/* base value of cycle counter */
static long cycle_counter_value;
#if defined(CHECK_MIGRATION)
/* (physical) processor number we are running */
static int processor_number = -1;
#endif


/*----------------------------------------------------------------------------*/
/* settings for  Alpha_21164/21264-True64-DEC/COMPAQ. */

#elif defined(ALPHA_DEC)

#define PCL_MAX_COUNTER 4         /* 21164 has 3 counters and a cycle counter */
                                  /* 21264 has 2 counters and a cycle counter */

/* although the hardware performance counters are 64 bit counters, the
   processor cycle counter has only 32 bits.
*/
#define COUNTER_SIZE_32
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/pfcntr.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <unistd.h>
#include <sys/processor.h>
#include <sys/resource.h>
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>
#include "machines/ALPHA_DEC/read_cycle_counter.h"
#include "machines/ALPHA_DEC/cpu_version.h"

/* shorts do best */
typedef short bit_mask_t;

/* initialized? */
static int initialized = 0;
/* pseudo-device for performance counters*/
static int fd = -1;
/* base value of cycle counter */
static long cycle_counter_value;
/* CPU number we are running on */
static long cpu_num;
/* what pfm counter we have selected */
static long counter_select;


/*----------------------------------------------------------------------------*/
/* settings for MIPS_R10000-IRIX-SGI. */

#elif defined(MIPS)

#define PCL_MAX_COUNTER 2

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/hwperfmacros.h>
#include <sys/hwperftypes.h>
#include <sys/procfs.h>

/* shorts do best */
typedef short bit_mask_t;

/* initialized? */
static int initialized = 0;
/* to check consistency */
static int generation_number;

/* counter device */
static int fd = -1;

/* initialized data structure to speed up things */
static hwperf_profevctrarg_t evctr_args_init;


/*----------------------------------------------------------------------------*/
/* settings for UltraSPARC-Solaris-SUN. */

#elif defined(SPARC)

#define PCL_MAX_COUNTER 3         /* 2 counters and one cycle counter */

#define COUNTER_SIZE_32           /* 32-bit performance counters */
#include <fcntl.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/types.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include "machines/SPARC/perfmon.h"

/* shorts do best */
typedef int bit_mask_t;

/* performance counter device */
static int fd = -1;
/* values of the PIC-registers when performance couting starts */
static unsigned long pic0_value, pic1_value;
/* base value of cycle counter */
static unsigned long long clock_value;


/*----------------------------------------------------------------------------*/
/* settings for PPC604e-AIX-IBM. */

#elif defined(POWERPC)

#define PCL_MAX_COUNTER 4

#define COUNTER_SIZE_32           /* 32-bit performance counters */
#include <stdlib.h>
#include "PMencode.h"
#include "PMprocess.h"
#include "PMevents.h"

static pm_int_regs ir;
static pm_event_info_list eil;
static struct PM_PROCESS_INIT pi;
static struct PM_PROCESS_CONTROL pc;
static struct PM_STATUS_CONTROL sc;
/* List of user events */
static pm_event_list uel;
/* List of initial values */
static pm_event_list uivl;

/* shorts do best */
typedef short bit_mask_t;


/*----------------------------------------------------------------------------*/
/* settings for Pentium/PPro/PII-Linux-* */

#elif defined(PENTIUM_LINUX)

#define PCL_MAX_COUNTER 3         /* 2 counter and cycle counter */

#define COUNTER_SIZE_32           /* 32-bit performance counters */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

/* model specific registers for Pentium and Pentium MMX */
#define PENTIUM_MSR_CTLEVSEL 0x11   /* MSR for control and event select */
#define PENTIUM_MSR_CNT0     0x12   /* MSR for counter 0 */
#define PENTIUM_MSR_CNT1     0x13   /* MSR for counter 1 */

/* model specific registers for PentiumPro, Pentium II, Pentium III */
#define PENTIUM2_MSR_EVNTSEL0 0x186 /* MSR for control and event select ctr0 */
#define PENTIUM2_MSR_EVNTSEL1 0x187 /* MSR for control and event select ctr1 */
#define PENTIUM2_MSR_PERFCTR0 0xc1  /* MSR for counter 0 */
#define PENTIUM2_MSR_PERFCTR1 0xc2  /* MSR for counter 1 */

/* shorts do best */
typedef short bit_mask_t;

/* initialized? */
static int initialized = 0;
/* msr device */
static int fd = -1;
/* base value of cycle counter */
static long long clock_value;
/* for Pentium MMX, Pentium Pro, Pentium II, and Pentium III 
   user level rdpmc is possible
   */
static int can_use_rdpmc = 0;


/*----------------------------------------------------------------------------*/
/* settings for Pentium/PPro/PII/PIII-NT-Microsoft */

#elif defined(PENTIUM_NT)

#define PCL_MAX_COUNTER 3         /* 2 counter and cycle counter */

#define COUNTER_SIZE_32           /* 32-bit performance counters */
/* not yet implemented */


/*----------------------------------------------------------------------------*/
/* else */

#else
#error "unsupported system"

#endif


#if defined(COUNTER_SIZE_32)
#include <signal.h>
#endif


/*                                                                            */
/*============================================================================*/
/* system independent variables with global scope                             */


/* PCL version number */
const char *PCL_version_number = "1.2";


/*                                                                            */
/*============================================================================*/
/* system independent variables with file-scope                               */


/*----------------------------------------------------------------------------*/
/* variables to generate result values.                                       */

/* performance counter values when starting nesting level */
static PCL_CNT_TYPE start_values[PCL_MAX_NESTING_LEVEL][PCL_MAX_COUNTER];

/* sum of values gotton so far from performance counters at a specific
   nesting level. E.g. on interrupts the current values of the performance
   counters are read out and added to the values on the current nesting level.
   */
static PCL_CNT_TYPE virtual_counters[PCL_MAX_NESTING_LEVEL][PCL_MAX_COUNTER];
static bit_mask_t virtual_counters_used_mask = 0;


/*----------------------------------------------------------------------------*/
/* these variables store arguments to PCLread for consistency checking on     */
/* subsequent PCLread/PCLstop calls.                                          */

/* count mode */
static unsigned int count_mode = PCL_MODE_USER;

/* number of requested events (must be same on each nesting level) */
static int event_count;

/* requested events (must be same on each nesting level) */
static int events[PCL_MAX_EVENT_PER_CALL];


/*----------------------------------------------------------------------------*/
/* misc. variables                                                            */

/* nesting level of PCL routines. Is incremented when PCLstart is called,
   and decremented when PCLstop is called.
 */
static int nesting_level = -1;

/* bit mask which counters are needed */
static bit_mask_t counter_used_mask = 0;

/* which events generate FP results */
static bit_mask_t fp_result_mask = 0;

/* MHz-rate of processor */
static int PCL_CNT_TYPE mhz_rate = -1;


#if defined(COUNTER_SIZE_32)           /* 32-bit performance counters */
/* old signal function */
static void (*old_signal_handler) (int);
#endif

/* names of events */
static const char *(eventname[PCL_MAX_EVENT]) =
{
  "PCL_L1CACHE_READ",
  "PCL_L1CACHE_WRITE",
  "PCL_L1CACHE_READWRITE",
  "PCL_L1CACHE_HIT",
  "PCL_L1CACHE_MISS",
  "PCL_L1DCACHE_READ",
  "PCL_L1DCACHE_WRITE",
  "PCL_L1DCACHE_READWRITE",
  "PCL_L1DCACHE_HIT",
  "PCL_L1DCACHE_MISS",
  "PCL_L1ICACHE_READ",
  "PCL_L1ICACHE_WRITE",
  "PCL_L1ICACHE_READWRITE",
  "PCL_L1ICACHE_HIT",
  "PCL_L1ICACHE_MISS",
  "PCL_L2CACHE_READ",
  "PCL_L2CACHE_WRITE",
  "PCL_L2CACHE_READWRITE",
  "PCL_L2CACHE_HIT",
  "PCL_L2CACHE_MISS",
  "PCL_L2DCACHE_READ",
  "PCL_L2DCACHE_WRITE",
  "PCL_L2DCACHE_READWRITE",
  "PCL_L2DCACHE_HIT",
  "PCL_L2DCACHE_MISS",
  "PCL_L2ICACHE_READ",
  "PCL_L2ICACHE_WRITE",
  "PCL_L2ICACHE_READWRITE",
  "PCL_L2ICACHE_HIT",
  "PCL_L2ICACHE_MISS",
  "PCL_TLB_HIT",
  "PCL_TLB_MISS",
  "PCL_ITLB_HIT",
  "PCL_ITLB_MISS",
  "PCL_DTLB_HIT",
  "PCL_DTLB_MISS",
  "PCL_CYCLES",
  "PCL_ELAPSED_CYCLES",
  "PCL_INTEGER_INSTR",
  "PCL_FP_INSTR",
  "PCL_LOAD_INSTR",
  "PCL_STORE_INSTR",
  "PCL_LOADSTORE_INSTR",
  "PCL_INSTR",
  "PCL_JUMP_SUCCESS",
  "PCL_JUMP_UNSUCCESS",
  "PCL_JUMP",
  "PCL_ATOMIC_SUCCESS",
  "PCL_ATOMIC_UNSUCCESS",
  "PCL_ATOMIC",
  "PCL_STALL_INTEGER",
  "PCL_STALL_FP",
  "PCL_STALL_JUMP",
  "PCL_STALL_LOAD",
  "PCL_STALL_STORE",
  "PCL_STALL",
  "PCL_MFLOPS",
  "PCL_IPC",
  "PCL_L1DCACHE_MISSRATE",
  "PCL_L2DCACHE_MISSRATE",
  "PCL_MEM_FP_RATIO"
};


/*----------------------------------------------------------------------------*/
/* For each event, there exists a table entry with the following components:.
   If the event is in some way available on the computer, "available" is
   set to 1.
   If the event is mapped directly to a low-level counter, "composed" is
   set to NOT_COMP and "data.command" contains the appropriate command word
   for the driver. Each entry not equal to -1 is a possible command word. As
   one event may be counted on different low-level counters, several command[i]
   may be different to -1.
   If the event is measured by a combination of some low-level counters,
   "composed" is set to COMPOSED and the rest of the data is stored in
   "composed_data". "composed_data.command" contains the command words and
   "composed.data.operation contains the reduction operation of the
   appropriate low-level counter values (e.g. (+1) * ctr_0 + (-1) * ctr_1).
   Example for non-composed entry:
   {PCL_AVAIL, NOT_COMP, {0x02, 0x03}}
   The event may be counted on counter 0 or counter 1, the command words for
   the low-level driver are 0x02 (for counter 0) and 0x03 (counter 1),
   respectively.
   Example for composed entry:
   {PCL_AVAIL, COMPOSED, {{0x0e, 0x0f}, {1, -1}}}
   The event is composed out of two events. The command words for
   the low-level driver are 0x0e (for counter 0) and 0x0f (counter 1),
   respectively. The resulting counter values (ctr_0, ctr_1) are connected in
   the form: (+1) * ctr_0 + (-1) * ctr_1
 */

/* additional bits to change operand type */
#define OP_FP_1         128u      /* change int operand 1 to floating point */
#define OP_FP_2         256u      /* change int operand 2 to floating point */
#define OP_FP_MASK      0x3f      /* mask out OP_FP_x */
#define OP_FP_TEST_MASK 0x180     /* mask to test if FP bit is set */

/* operations (might be combined with OP_FP_x) */
/* operations are strictly applied left to right, i.e. all operations on
   counter 0 (if used) followed by all operations on counter 1 (if used) etc. */
#define OP_ADD         1u      /* int add value */
#define OP_SUB         2u      /* int subtract value */
#define OP_FP_STORE    3u      /* floating point store */
#define OP_FP_ADD      4u      /* floating point add */
#define OP_FP_MUL      5u      /* floating point multiply */
#define OP_FP_DIV      6u      /* floating point divide */
#define OP_FP_DIV_MHZ  7u      /* floating point divide by value and multiply with MHz */
#define OP_FP_REZ      8u      /* add 1.0/value */
#define OP_FP_REZ_MHZ  9u      /* add MHz/value */

#define OP_NO         (~0u)    /* no entry */

#define COMPOSED 1             /* event composed */
#define NOT_COMP 0             /* event not composed */


typedef struct
  {
    unsigned int available:1;	          /* event is available in HW/driver? */
    unsigned int composed:1;	          /* event is composed? */
    union
      {
	struct
	  {
	    long command[PCL_MAX_COUNTER];      /* command values */
	    unsigned int operation[PCL_MAX_COUNTER];   /* combine oper. */
	  }
	composed_data;
	long command[PCL_MAX_COUNTER];        /* command value */
      }
    data;
  }
event_descr_t;


/* table of event entries (see above):
static event_descr_t event_table[PCL_MAX_EVENT];
*/


/*----------------------------------------------------------------------------*/
/* Inverse table entries. Given the results in ctr_0,...,ctr_n, this table
   specifies, how the PCL result values are constructed and where they are
   stored. Each hardware counter has such an entry.
 */

typedef struct
  {
    PCL_CNT_TYPE result_value;	       /* result value of low-level driver */
    long command;                      /* command to activate low-level cntr */
    short result_index;		       /* contributes to what user result */
    short event;                       /* used on what event */
    unsigned short operation;          /* combine operation */
  }
result_descr_t;


/* for each low-level counter one entry */
static result_descr_t results[PCL_MAX_COUNTER];


/*----------------------------------------------------------------------------*/

/* marker in table if this event is not available/available */
#define PCL_UNAVAIL      0
#define PCL_AVAIL        1


/*----------------------------------------------------------------------------*/

#if defined(ALPHA_CRAY)

/* allowed count modes */
static unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1,-1},{0,0,0,0}}}

/* entries are control codes for hardware performance counter.
   sel0-sel2 contain event selectors for counter0 - counter2. For counter1
   and counter2, CBOX events are decoded as event selectors larger than 0x0e.
   The CBOX value is then computed as (value - 0x0f).
*/
static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xE,-1,-1},{OP_NO,OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,0xE,0x6,-1},{-1,OP_ADD,OP_SUB,-1}}}},/* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x6,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xD,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,0xD,0x3,-1},{-1,OP_ADD,OP_SUB,-1}}}},/* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x3,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x10,-1,-1},{-1,OP_ADD,-1,-1}}}},/* PCL_L2CACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x11,-1,-1},{-1,OP_ADD,-1,-1}}}},/* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xF,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,0xF,0xF,-1},{-1,OP_ADD,OP_SUB,-1}}}},/* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0xF,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x5,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x7,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0,-1,0xc,-1},{OP_ADD,-1,OP_ADD,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,0x0},{-1,-1,-1,OP_ADD}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{-1,0x9,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xa,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xb,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,0xc,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x1,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL, PCL_INVALID},                      /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x2,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL, PCL_INVALID},                      /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0xe,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{-1,0xa,0xc,-1},{-1,OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0x1,-1,0xc,-1},{OP_FP_2|OP_FP_STORE,-1,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{-1,0xe,0x6,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                     /* PCL_MEM_FP_RATIO */
};

/*----------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

/* allowed count modes */
static unsigned int allowed_count_mode =  PCL_MODE_USER_SYSTEM;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1,-1},{0,0,0,0}}}

/* entries are control codes for hardware performance counter.
   Distinguish at runtime between 21164 and 21264 CPUs 
 */
static event_descr_t *event_table = NULL;

/* entries are control codes for hardware performance counter */
static event_descr_t event_table_21164[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,PF5_MUX2_C_DCACHE_MISSES,-1},{-1,OP_ADD,OP_SUB,-1}}}}, /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_DCACHE_MISSES,-1},{-1,-1,OP_ADD,-1}}}},	/* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_ICACHE_ACCESS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_ICACHE_ACCESS,PF5_MUX2_C_ICACHE_MISSES,-1},{-1,OP_ADD,OP_SUB,-1}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_ICACHE_MISSES,-1},{-1,-1,OP_ADD,-1}}}},	/* PCL_L1ICACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_READ,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_WRITE,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,PF5_MUX2_C_SCACHE_MISS,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_SCACHE_MISS,-1},{-1,-1,OP_ADD,-1}}}},	/* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_ITB_MISS,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_DTB_MISS,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PF5_MUX0_CYCLES,-1,PF5_MUX2_C_CYCLES,-1},{OP_ADD,-1,OP_ADD,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,-1,0x00},{-1,-1,-1,OP_ADD}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_INTEGER_OPERATE,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_FP_INSTRUCTIONS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_LOAD_INSTRUCTIONS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF5_MUX1_C_STORE_INSTRUCTIONS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PF5_MUX0_ISSUES,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_BRANCH_MISPREDICTS,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PF5_MUX2_C_LOAD_LOCKED,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_FP_INSTRUCTIONS,PF5_MUX2_C_CYCLES,-1},{-1,OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{PF5_MUX0_ISSUES,-1,PF5_MUX2_C_CYCLES,-1},{OP_FP_2|OP_FP_STORE,-1,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,PF5_MUX2_C_DCACHE_MISSES,-1},{-1,OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                     /* PCL_MEM_FP_RATIO */
};


/* logical constants for 21264 systems are available in <sys/pfcntr.h> */
/* entries are control codes for hardware performance counter */
#if !defined(OS_HAS_SUPPORT_FOR_21264)
/* older versions of Digital Unix have no support for 21264 */
static event_descr_t event_table_21264[PCL_MAX_EVENT];
#else
static event_descr_t event_table_21264[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                  /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                  /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                  /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF6_MUX1_RET_ITB_MISSES,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PF6_MUX0_CYCLES,PF6_MUX1_CYCLES,-1,-1},{OP_ADD,OP_ADD,-1,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_INTEGER_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PF6_MUX0_RET_INSTRUCTIONS,-1,-1,-1},{OP_ADD,-1,-1,-1}}}},  /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP_SUCCESS */
#if 0
  /* according to the documentation of pfm the event for mispredicted branches
     does actually count the number of cycles.
     */
  {PCL_AVAIL,NOT_COMP,{{{-1,PF6_MUX1_RET_BRANCH_MISPREDICTS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_JUMP_UNSUCCESS */
#else
  {PCL_UNAVAIL ,PCL_INVALID},                     /* PCL_JUMP_UNSUCCESS */
#endif
  {PCL_AVAIL,NOT_COMP,{{{-1,PF6_MUX1_RET_COND_BRANCHES,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL ,PCL_INVALID},                     /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                      /* PCL_MEM_FP_RATIO */
};
#endif


/*----------------------------------------------------------------------------*/

#elif defined(MIPS)

/* allowed count modes */
static unsigned int allowed_count_mode = PCL_MODE_USER;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1},{0,0}}}

/* entries are control codes for hardware performance counter */
static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{9,25},{OP_ADD,OP_ADD}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,25},{-1,OP_ADD}}}},    /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{9,-1},{OP_ADD,-1}}}},     /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{10,26},{OP_ADD,OP_ADD}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,26},{-1,OP_ADD}}}},    /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_L2ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{10,-1},{OP_ADD,-1}}}},    /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_TLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,23},{-1,OP_ADD}}}},    /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0,16},{OP_ADD,OP_ADD}}}},      /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,21},{-1,OP_ADD}}}},    /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,18},{-1,OP_ADD}}}},    /* PCL_LOAD_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{-1,19},{-1,OP_ADD}}}},    /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{15,17},{OP_ADD,OP_ADD}}}}, /* PCL_INSTR */
  {PCL_AVAIL,COMPOSED,{{{6,24},{OP_ADD,OP_SUB}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,24},{-1,OP_ADD}}}},    /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{6,-1},{OP_ADD,-1}}}},     /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,20},{-1,OP_ADD}}}},    /* PCL_ATOMIC_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{5,-1},{OP_ADD,-1}}}},     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_AVAIL,COMPOSED,{{{5,20},{OP_ADD,OP_ADD}}}},      /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	              /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0,21},{OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL}}}}, /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                 /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                  /* PCL_MEM_FP_RATIO */
};

/*----------------------------------------------------------------------------*/


#elif defined(SPARC)

/* allowed count modes */
static unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1},{OP_NO,OP_NO,OP_NO}}}

static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_DC_READ,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_DC_WRITE,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_EC_READ_HIT,-1,-1},{OP_ADD,OP_NO,OP_NO}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_IC_REF,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PCR_S1_IC_HIT,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{PCR_S0_IC_REF,PCR_S1_IC_HIT,-1},{OP_ADD,OP_SUB,OP_NO}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_EC_REF,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,NOT_COMP,{{{-1,PCR_S1_EC_HIT,-1},{OP_NO,OP_ADD,OP_NO}}}},  /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,PCR_S1_EC_WRITEBACK,-1},{OP_NO,OP_ADD,OP_NO}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_CYCLE_CNT,PCR_S1_CYCLE_CNT,-1},{OP_ADD,OP_ADD,OP_NO}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{OP_NO,OP_NO,OP_ADD}}}},        /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_INTEGER_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PCR_S0_INSTR_CNT,-1,-1},{OP_ADD,OP_NO,OP_NO}}}},  /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{PCR_S0_INSTR_CNT,PCR_S1_CYCLE_CNT,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,OP_NO}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                      /* PCL_L1DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{PCR_S0_EC_READ_HIT,PCR_S1_EC_WRITEBACK,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,OP_NO}}}}, /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                       /* PCL_MEM_FP_RATIO */
};


/*----------------------------------------------------------------------------*/

#elif defined(POWERPC)

/* allowed count modes */
static unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* bringing PCL and PMapi together, you need x different events (countable on
   counter x only) to actually count nothing on x counters
*/
static long default_event[PCL_MAX_COUNTER] =
 {
   PM_IC_MISS,
   PM_DC_MISS,
   PM_FPU_WT,
   PM_BTAC_HITS
 };

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1,-1},{0,0,0,0}}}

static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,COMPOSED,{{{PM_IC_MISS,PM_DC_MISS,-1,-1},{OP_ADD,OP_ADD,-1,-1}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,PM_DC_MISS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{PM_IC_MISS,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_TLB_HIT */
  {PCL_AVAIL,COMPOSED,{{{PM_DTLB_MISS,PM_ITLB_MISS,-1,-1},{OP_ADD,OP_ADD,-1,-1}}}}, /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{-1,PM_ITLB_MISS,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{PM_DTLB_MISS,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{PM_CYC,PM_CYC,PM_CYC,PM_CYC},{OP_ADD,OP_ADD,OP_ADD,OP_ADD}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ELAPSED_CYCLES */
  {PCL_AVAIL,NOT_COMP,{{{PM_FXU_CMPL,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PM_FPU_CMPL,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PM_LS_EXEC,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{PM_INST_DISP,-1,-1,-1},{OP_ADD,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{-1,PM_BR_CMPL,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{-1,PM_RESRV_CMPL,-1,-1},{-1,OP_ADD,-1,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_INTEGER */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PM_FPU_WT,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_STALL_FP */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,PM_BRU_WT,-1},{-1,-1,OP_ADD,-1}}}}, /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{PM_FPU_CMPL,PM_CYC,-1,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{PM_INST_DISP,PM_CYC,-1,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1,-1}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}                 /* PCL_MEM_FP_RATIO */
};

/*----------------------------------------------------------------------------*/


#elif defined(PENTIUM_LINUX) || defined(PENTIUM_NT)

/* allowed count modes */
static unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* entries are control codes for hardware performance counter.
   Distinguish at runtime between Pentium, Pentium MMX,
   and Pentium Pro/Pentium II/Pentium III CPUs 
 */
static event_descr_t *event_table = NULL;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1},{0,0,0}}}

static event_descr_t event_table_pentium[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x00,0x00,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x01,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x00,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0C,0x0C,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0E,0x0E,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0D,0x0D,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x02,0x02,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{-1,-1,OP_ADD}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x22,0x22,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x16,0x16,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0x12,0x12,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{0x1A,0x1A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{0x19,0x19,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_MFLOPS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x28,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0x28,0x22,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}} /* PCL_MEM_FP_RATIO */
};


static event_descr_t event_table_pentium_mmx[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x00,0x00,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x01,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x00,0x01,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x0C,0x0C,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0E,0x0E,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x0D,0x0D,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x02,0x02,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x30,-1,-1},{OP_ADD,-1,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{-1,-1,OP_ADD}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x22,0x22,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x28,0x28,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0x16,0x16,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0x12,0x12,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_AVAIL,NOT_COMP,{{{0x32,0x32,-1},{OP_ADD,OP_ADD,-1}}}},  /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,NOT_COMP,{{{0x1A,0x1A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,NOT_COMP,{{{0x19,0x19,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0x30,0x22,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ_MHZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0x30,0x16,-1},{OP_FP_1|OP_FP_2|OP_FP_REZ,OP_FP_2|OP_FP_MUL,-1}}}}, /* PCL_IPC */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x28,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_AVAIL,COMPOSED,{{{0x28,0x22,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}} /* PCL_MEM_FP_RATIO */
};


static event_descr_t event_table_pentium2[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x45,0x45,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x80,0x80,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x81,0x81,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,NOT_COMP,{{{0x2E,0x2E,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x24,0x24,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2CACHE_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x29,0x29,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2DCACHE_READ */
  {PCL_AVAIL,NOT_COMP,{{{0x2A,0x2A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2DCACHE_WRITE */
  {PCL_AVAIL,COMPOSED,{{{0x29,0x2A,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,NOT_COMP,{{{0x85,0x85,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_MISS */
  {PCL_AVAIL,NOT_COMP,{{{0x79,0x79,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_CYCLES */
  /* special case with dummy entry */
  {PCL_AVAIL,NOT_COMP,{{{-1,-1,0x00},{-1,-1,OP_ADD}}}},  /* PCL_ELAPSED_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC1,-1,-1},{OP_ADD,-1,-1}}}},  /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC0,0xC0,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,NOT_COMP,{{{0xC9,0xC9,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC5,0xC5,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,NOT_COMP,{{{0xC4,0xC4,-1},{OP_ADD,OP_ADD,-1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_AVAIL,NOT_COMP,{{{0xA2,0xA2,-1},{OP_ADD,OP_ADD,-1}}}},  /* PCL_STALL */
  {PCL_AVAIL,COMPOSED,{{{0xC1,0x79,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV_MHZ,-1}}}}, /* PCL_MFLOPS */
  {PCL_AVAIL,COMPOSED,{{{0xC0,0x79,-1},{OP_FP_2|OP_FP_STORE,OP_FP_2|OP_FP_DIV,-1}}}}, /* PCL_IPC */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L2DCACHE_MISSRATE */
  {PCL_UNAVAIL, PCL_INVALID}           /* PCL_MEM_FP_RATIO */
};


/*----------------------------------------------------------------------------*/

#endif


/*                                                                            */
/*============================================================================*/
/* forward declarations */

#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)

static void
interrupt_handling(
		    int sig
);

#endif


/*                                                                            */
/*============================================================================*/
/* Some functions specific to the 21x64 processor line.                       */


#if defined(ALPHA_DEC)

/* We determine what processor type the program is running on and therefore
   what events are supported. 
   Sets "event_table" accordingly to get the right register settings.
   Additionally, the pseudo device is opened and the process gets bound
   to the processor.
   Returns PCL_SUCCESS if processor type is supported.
*/
static int
initialize_21x64(void)
{
  long cpu;


  /* get processor type */
  cpu = cpu_implementation_version();
  if(cpu == 1)
    event_table = event_table_21164;
  else if(cpu == 2)
    event_table = event_table_21264;
  else
    /* something we can't handle yet */
    return PCL_NOT_SUPPORTED;

  /* Try to open a channel to the pfcntr driver (if not already opened) */
  if(fd == -1)
    {
      unsigned long cpu_mask;
      
      /* get actual CPU we are running on */
      if(getsysinfo(GSI_CURRENT_CPU, &cpu_num, sizeof(cpu_num),
		    NULL, NULL, NULL) < 1)
	return PCL_FAILURE;
      
      /* bind process to processor */
      cpu_mask = 1 << cpu_num;
      if (bind_to_cpu(getpid(), cpu_mask, BIND_NO_INHERIT) == -1)
	return PCL_FAILURE;

      /* open device (only for this processor) */
      fd = open("/dev/pfcntr", O_RDONLY | PCNTOPENONE,
		S_IRUSR | S_IRGRP | S_IROTH);
      if (fd == -1)
	return PCL_FAILURE;
    }


  /* everything ok */
  return PCL_SUCCESS;
}

#endif


/*                                                                            */
/*============================================================================*/
/* Some functions specific to the R10000 processor line (actually specific to */
/* the /proc interface of the IRIX operating system).                         */


#if defined(MIPS)

static int
initialize_r1x000(void)

{
  char pfile[12];
  pid_t pid;
  int cntr;


  /* open the /proc-interface for performance counter access */
  if((pid = getpid()) == -1)
    return PCL_FAILURE;
  sprintf(pfile, "/proc/%05d", (int)pid);
  if((fd = open(pfile, O_RDWR)) == -1)
    return PCL_FAILURE;

  /* Now initialize a data structure used in counter descriptions.
     Ever seen such a few things to set up in such a complicated interface?
  */
  evctr_args_init.hwp_ovflw_sig = 0;
  for(cntr = 0; cntr < HWPERF_EVENTMAX; ++cntr)
    {
      /* overflow frequency */
      evctr_args_init.hwp_ovflw_freq[cntr] = 0;
      /* don't use counter */
      evctr_args_init.hwp_evctrargs.hwp_evctrl[cntr].hwperf_spec = 0;
    }

  return PCL_SUCCESS;
}

#endif


/*                                                                            */
/*============================================================================*/
/* Some functions specific to the Intel processor line.                       */


#if defined(PENTIUM_LINUX)


/* fast assembler code to access the cycle counter (TSC) */
#define rdtsc(valptr) asm volatile ("rdtsc" : \
"=a"(*(unsigned int *)valptr), "=d"(*(((unsigned int *)valptr)+1)))

/* fast assembler code to read out the performance counters (only possible
   on Pentium MMX, Pentium Pro, Pentium II, Pentium III
   and if allowed by setting a special bit in kernel mode) */
static inline long long
rdpmc(int reg)
{
  unsigned int tmp[2];

  asm volatile("rdpmc" : "=a" (tmp[0]), "=d" (tmp[1]) : "c" (reg));
  return ((long long)(tmp[1] & 0xff) << 32) | tmp[0];
}


/* We determine if the processor supports performance counters (special MSR)
   and what events are supported. Therefore we use the CPUID instruction to
   test certain features.
   Sets "event_table" accordingly to get the right register settings.
   Returns PCL_SUCCESS if processor type is supported.
*/
static int
initialize_pentium(void)
{
  unsigned int signature, feature_flags, family;


  /* get all processor type information with the CPUID instruction
     Calling CPUID with EAX=1 returns in EAX version information (type, family,
     model, stepping ID) and in EDX feature information.
     The information in EAX is:
     bit  0-3 : stepping
     bit  4-7 : model
     bit  8-11: family
                0100 = i486
                0101 = Pentium
                0110 = Pentium Pro or Pentium II or Pentium III
     bit 12-13: processor type
     The information in EDX is (only relevant information shown here):
     bit 4 : Time Stamp Counter supported
     bit 5 : Model Specific Registers implemented with RDMSR, WRMSR instructions
     bit 23: MMX supported
   */
  asm("cpuid"
      : "=a"(signature), "=d"(feature_flags)
      : "a"(1)
      : "%eax", "%ebx", "%ecx", "%edx");

  /* check for existence of time stamp counter (TSC) */
  if((((feature_flags >> 4) & 0x1) != 1)
     /* and check for existence of model specific registers (MSR) */
     || (((feature_flags >> 5) & 0x1) != 1))
    /* something old without support for performance counters */
    return PCL_NOT_SUPPORTED;

  /* preset value */
  can_use_rdpmc = 1;

  /* now test what processor family we have to choose the right event table */
  family = (signature >> 8) & 0xf;
  if (family == 5)
    {
      /* Pentium processor; test now for Pentium/MMX */
      if(((feature_flags >> 23) & 0x1) == 0)
	{
	  /* Pentium without MMX */
	  event_table = event_table_pentium;
	  /* no user-level rdpmc instruction possible */
	  can_use_rdpmc = 0;
	}
      else
	/* Pentium MMX */
	event_table = event_table_pentium_mmx;
    }

  else if (family == 6)
    /* Pentium Pro, Pentium II, Pentium III */
    event_table = event_table_pentium2;

  else
    /* something we can't handle yet */
    return PCL_NOT_SUPPORTED;

  /* open msr device */
  if(fd == -1)
    {
      /* open msr device for read/write */
      if((fd = open("/dev/msr", O_RDWR)) == -1)
	return PCL_FAILURE;
    }


  return PCL_SUCCESS;
}

#endif


/*                                                                            */
/*============================================================================*/
/* Start hardware couters.                                                    */
/* This includes setting up the intervall timer and the appropriate signal    */
/* handler to catch timer interrupts.                                         */


static int
hw_counter_start(
		void
)

{
  /* 32-bit performance counters */
#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)
  static struct itimerval itimer =
  {{PCL_INTERRUPT_INTERVALL / 1000, PCL_INTERRUPT_INTERVALL % 1000},
   {PCL_INTERRUPT_INTERVALL / 1000, PCL_INTERRUPT_INTERVALL % 1000}};
  

  /* set up signal handler */
  if(old_signal_handler != interrupt_handling)
    {
      old_signal_handler = signal(ITIMER_SIGNAL, interrupt_handling);
      if (old_signal_handler == SIG_ERR)
	return PCL_FAILURE;
    }

  /* set up timer to deliver signal */
  if(setitimer(ITIMER_MODE, &itimer, NULL) != 0)
    return PCL_FAILURE;
#endif


  /*--------------------------------------------------------------------------*/
  /* system specific actions */

#if defined(ALPHA_CRAY)

  {
    int sel0, sel1, sel2;	       /* counter selectors */
    unsigned int mode;


    /* initialize selectors with default count events (cycles,
       non_issue_cycles, long_stalls) */
    sel0 = sel1 = sel2 = 0;

    /* Extract selected events */
    if(IS_USED(counter_used_mask, 0))
      sel0 = results[0].command;
    if(IS_USED(counter_used_mask, 1))
      sel1 = results[1].command;
    if(IS_USED(counter_used_mask, 2))
      sel2 = results[2].command;

    /* select count mode */
    if(count_mode == PCL_MODE_USER)
      mode = 0x01;
    else if(count_mode == PCL_MODE_SYSTEM)
      mode = 0x02;
    else
      mode = 0x03;

    /* get actual processor cycle counter value (to measure elapsed cycles) */
    cycle_counter_value = ev5_read_cycle_counter();

    /* Clear and start the counters */
    return (ev5_perfctr_start(&sel0, &sel1, &sel2, mode) == 0)
      ? PCL_SUCCESS
      : PCL_FAILURE;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

  {
    long arg, rc, event_select, int_frequency;
    union pmctrs_ev5 pm_ev5;
#if defined(OS_HAS_SUPPORT_FOR_21264)
    union pmctrs_ev6 pm_ev6;
#endif


    if(event_table == event_table_21164)
      {
	/* 21164 */
	/* initialize selectors */
	event_select = counter_select = int_frequency = 0;

	/* extract selected events */
	if(IS_USED(counter_used_mask, 0))
	  {
	    event_select = results[0].command;
	    counter_select = PF5_SEL_COUNTER_0;
	    int_frequency = PF5_C0_INT_EVERY_65536;
	  }
	if(IS_USED(counter_used_mask, 1))
	  {
	    event_select |= results[1].command;
	    counter_select |= PF5_SEL_COUNTER_1;
	    int_frequency |= PF5_C1_INT_EVERY_65536;
	  }
	if(IS_USED(counter_used_mask, 2))
	  {
	    event_select |= results[2].command;
	    counter_select |= PF5_SEL_COUNTER_2;
	    int_frequency |= PF5_C2_INT_EVERY_16384;
	  }

	/* select items (= counters) */
	arg = PFM_COUNTERS;
	rc = ioctl(fd, PCNTSETITEMS, &arg);
	if (rc < 0)
	  return PCL_FAILURE;

	/* clear driver's counters */
	rc = ioctl(fd, PCNTCLEARCNT);
	if (rc < 0)
	  return PCL_FAILURE;

	/* set contexts in which to count */
	arg = PF5_CTXT_THIS_PROC;
	/* can anybody explain to me why the constants PF5_CTXT_... have a
	   meaning different to that what their name suggests?
	   If you want an explanation for the settings, check how it is done on
	   the CRAY.
	   */
	if(count_mode == PCL_MODE_USER)
	  /* user mode only: 0x40000200 */
	  arg |= PF5_CTXT_USER_AND_KERNEL;
	else if(count_mode == PCL_MODE_SYSTEM)
	  /* kernel mode only: 0x40000100 */
	  arg |= PF5_CTXT_PAL;
	else
	  /* user or kernel mode: 0x00000000 */
	  arg |= PF5_CTXT_ALL_MODES;
	rc = ioctl(fd, PCNT5CTXTS, &arg);
	if (rc < 0)
	  return PCL_FAILURE;

	/* select events */
	arg = event_select;
	rc = ioctl(fd, PCNT5MUX, &arg);
	if (rc < 0)
	  return PCL_FAILURE;
	
	/* select interrupt frequency */
	arg = int_frequency;
	rc = ioctl(fd, PCNT5FREQ, &arg);
	if (rc < 0)
	  return PCL_FAILURE;
	
	/* zero and restart selected counters */
	pm_ev5.pmctrs_ev5_cpu = cpu_num;
	pm_ev5.pmctrs_ev5_select = counter_select;
	rc = ioctl(fd, PCNT5RESTART, &pm_ev5.pmctrs_ev5_long);
	if (rc < 0)
	  return PCL_FAILURE;
      }

#if defined(OS_HAS_SUPPORT_FOR_21264)
    else
      {
	/* 21264 */
	/* initialize selectors */
	event_select = counter_select = 0;

	/* extract selected events */
	if(IS_USED(counter_used_mask, 0))
	  {
	    event_select = results[0].command;
	    counter_select = PF6_SEL_COUNTER_0;
	  }
	if(IS_USED(counter_used_mask, 1))
	  {
	    event_select |= results[1].command;
	    counter_select |= PF6_SEL_COUNTER_1;
	  }

	/* select items (= counters) */
	arg = PFM_COUNTERS;
	rc = ioctl(fd, PCNTSETITEMS, &arg);
	if (rc < 0)
	  return PCL_FAILURE;

	/* clear driver's counters */
	rc = ioctl(fd, PCNTCLEARCNT);
	if (rc < 0)
	  return PCL_FAILURE;

#if 0   /* doesn't work as I think it should do */
	/* for 21264-based systems, pfm doesn't support the distinction
	   between user and system mode. What we do here is to count only
	   those processes which have performance counting bit enabled. */
	rc = ioctl(fd, PCNTLOGSELECT);
	if (rc < 0)
	  return PCL_FAILURE;
#endif

	/* select events */
	arg = event_select;
	rc = ioctl(fd, PCNT6MUX, &arg);
	if (rc < 0)
	  return PCL_FAILURE;

	/* zero and restart selected counters */
	pm_ev6.pmctrs_ev6_cpu = cpu_num;
	pm_ev6.pmctrs_ev6_select = counter_select;
	rc = ioctl(fd, PCNT6RESTART, &pm_ev6.pmctrs_ev6_long);
	if (rc < 0)
	  return PCL_FAILURE;
      }
#endif


    /* get actual processor cycle counter value (to measure elapsed cycles) */
    cycle_counter_value = read_cycle_counter();

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)

  {
    int sel0, sel1;
    hwperf_profevctrarg_t evctr_args;


    /* check correct counting mode */
    if(count_mode != PCL_MODE_USER)
      return PCL_MODE_NOT_SUPPORTED;

    /* initialize counter descriptors */
    memcpy(&evctr_args, &evctr_args_init, sizeof(evctr_args));
    
    if(IS_USED(counter_used_mask, 0))
      {
	/* counter 0 used */
	sel0 = results[0].command;
	/* event to be counted */
	evctr_args.hwp_evctrargs.hwp_evctrl[sel0].hwperf_creg.hwp_ev = sel0;
	/* overflow interrupt enable */
	evctr_args.hwp_evctrargs.hwp_evctrl[sel0].hwperf_creg.hwp_ie = 1;
	/* user/kernel mode */
	evctr_args.hwp_evctrargs.hwp_evctrl[sel0].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
      }

    if(IS_USED(counter_used_mask, 1))
      {
	/* counter 1 */
	sel1 = results[1].command;
	/* event to be counted */
	evctr_args.hwp_evctrargs.hwp_evctrl[sel1].hwperf_creg.hwp_ev = sel1;
	/* overflow interrupt enable */
	evctr_args.hwp_evctrargs.hwp_evctrl[sel1].hwperf_creg.hwp_ie = 1;
	/* user/kernel mode */
	evctr_args.hwp_evctrargs.hwp_evctrl[sel1].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
      }

    /* acquire counters */
    generation_number = ioctl(fd, PIOCENEVCTRS, (void *)&evctr_args);
    if(generation_number < 0)
      return PCL_FAILURE;
    else
      return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

  {
    int sel0, sel1;
    unsigned long long set_count, val;
    processorid_t proc;


    /* initialize selectors with default count events (cycles, cycles) */
    sel0 = sel1 = 0;

    /* Extract selected events */
    if(IS_USED(counter_used_mask, 0))
      sel0 = results[0].command;
    if(IS_USED(counter_used_mask, 1))
      sel1 = results[1].command;

    /* if counter device is not already opened, open device */
    if (fd == -1)
      {
	/* Try to open a channel to the perfmon driver */
	fd = open("/dev/perfmon", O_RDONLY);
	if (fd == -1)
	  return PCL_FAILURE;

	/* query on what processor we are running */
	if (processor_bind(P_LWPID, P_MYID, PBIND_QUERY, &proc) == -1)
	  return PCL_FAILURE;

	/* if not assigned take first CPU available */
	if (proc == -1)
	  for(proc = 0; p_online(proc, P_STATUS) != P_ONLINE; ++proc)
	    ;

	/* bind process to processor */
	if (processor_bind(P_LWPID, P_MYID, proc, NULL) == -1)
	  return PCL_FAILURE;
      }

    /* Set PCR to accumulate desired events */
    set_count = sel0 | sel1;
    if(count_mode == PCL_MODE_USER)
      /* user mode only */
      set_count |= PCR_USER_TRACE;
    else if(count_mode == PCL_MODE_SYSTEM)
      /* kernel mode only */
      set_count |= PCR_SYS_TRACE;
    else
      /* user or kernel mode */
      set_count |= (PCR_SYS_TRACE | PCR_USER_TRACE);
    if(ioctl(fd, PERFMON_SETPCR, &set_count) < 0)
      return PCL_FAILURE;

    /* Clear and start counter */
    clr_pic();
    cpu_sync();

    /* read out counter values (used as base value) */
    val = get_pic();
    pic0_value = val & 0xffffffff;
    pic1_value = val >> 32;

    /* get actual clock value (to measure cycles) */
    clock_value = get_tick();


    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  {
    int i;
    struct PM_ENCODING_CONTROL uic;
    struct PM_MACHINE_CONTROL mc;
    union PM_RETN retn;


    /* Initialize control blocks to the default value, 0, for each field.  */
    /* This routine clears all fields in uic,sc,pi,pc to present the defaults.*/
    memset(&uic, 0, sizeof(uic));
    memset(&sc, 0, sizeof(sc));
    memset(&pi, 0, sizeof(pi));
    memset(&pc, 0, sizeof(pc));


    /* request encoding */
    /* Revision */
    uic.revision = PM_ENCODING_CONTROL_REV;
    /* list of requested events */
    uic.uel = &uel;
    /* corresponding list for requested events */
    uic.ueivl = &uivl;
    /* No event chain history control list */
    uic.uechl = NULL;
    /* No event counter control list */
    uic.uecntl = NULL;
    /* No event group control list */
    uic.uegrpl = NULL;

    /* Count in problem state */
    if(count_mode == PCL_MODE_USER)
      uic.c_fields.b.MACH_STATE = MACH_STATE_PROB;
    else if(count_mode == PCL_MODE_SYSTEM)
      uic.c_fields.b.MACH_STATE = MACH_STATE_SUPERV;
    else
      uic.c_fields.b.MACH_STATE = MACH_STATE_ANY;

    /* Count regardless of the state of the MM bit in the MSR */
    uic.c_fields.b.PMBIT = PMBIT_ANY;

    /* event list encoding */
    /* Count PCL_MAX_COUNTER events */
    uel.n = PCL_MAX_COUNTER;
    for(i = 0; i < PCL_MAX_COUNTER; ++i)
      {
	/* not PM_NOTHING! otherwise we get the wrong order of results */
	uel.e[i] = (IS_USED(counter_used_mask, i))
	  ? results[i].command : default_event[i];
      }

    /* start encoding */
    retn.w = PMencode(&uic, &mc);


    /* process init */
    /* Set revision field */
    pi.revision = PM_PROCESS_INIT_REV;
    /* Post caller on each interrupt */
    pi.pc_flags.b.POSTMODE = PM_POSTMODE_INTERRUPT;
    /* Automatically count next set of events on each interrupt */
    pi.pc_flags.b.AUTOMODE = PM_AUTOMODE_NEXT;
    /* Don't start counting on PMprocessInit call */
    pi.pc_flags.b.INITNOSTART = PM_INIT_NOSTART;
    /* Continue counting forever */
    pi.tp = 1;


    /* Initialize for processing */
    retn.w = PMprocessInit(&pi, &mc);
    if (retn.w != 0)
      return PCL_FAILURE;


    /* set processor control */
    /* Set revision field */
    pc.revision = PM_PROCESS_CONTROL_REV;
    /* Start counting command */
    pc.command = PM_COUNT_START;


    /* status control */
    /* set revision field */
    sc.revision = PM_STATUS_CONTROL_REV;
    /* control command */
    sc.command = PM_STATUS_OPTION_ADJUST;
    /* No list of requested events */
    sc.elp = NULL;
    sc.eip = &eil;
    sc.sip = &ir;
    sc.tip = NULL;


    /* Start counting */
    retn.w = PMprocess(&pc);
    if (retn.w != 0)
      return PCL_FAILURE;

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)

  {
    int sel0, sel1, cpl;
    unsigned long long clear;


    /* initialize selectors with default count events (something existing) */
    /* FP-instr exists for all Pentiums */
    sel0 = sel1 = event_table[PCL_FP_INSTR].data.command[0];

    /* Extract selected events */
    if(IS_USED(counter_used_mask, 0))
      sel0 = results[0].command;
    if(IS_USED(counter_used_mask, 1))
      sel1 = results[1].command;

    /* check count mode */
    if(count_mode == PCL_MODE_USER)
      /* count only in user mode */
      cpl = 2;
    else if(count_mode == PCL_MODE_SYSTEM)
      /* count only system mode */
      cpl = 1;
    else
      /* count in system or user mode */
      cpl = 3;


    if ((event_table == event_table_pentium)
	|| (event_table == event_table_pentium_mmx))
      {
	/* sub-fields in the Control and Event Select Register (CESR)
	 *     CC0, CC1: CPL Level to Monitor, possibilities are
	 *               000 = Count Nothing
	 *               001 = Count Event while CPL = 0,1,2
	 *               010 = Count Event while CPL = 3
	 *               011 = Count Event while CPL = 0,1,2,3
	 *               100 = Count Nothing
	 *               101 = Count Clocks while CPL = 0,1,2
	 *               110 = Count Clocks while CPL = 3
	 *               111 = Count Clocks while CPL = 0,1,2,3
	 *               Here we only use 000, 001, 010 and 011.
	 */
	union
	{
	  struct pentium_cesr
	  {
	    unsigned int es0:6;        /* event select counter 0 */
	    unsigned int cc0:3;        /* counter control 0 (see above) */
	    unsigned int pc0:1;        /* pin control bit 0
					  0=event increment, 1=event overflow */
	    unsigned int reserved1:6;  /* reserved */
	    unsigned int es1:6;        /* event select counter 1 */
	    unsigned int cc1:3;        /* counter control 1 (see cc0) */
	    unsigned int pc1:1;        /* pin control bit 1 */
	    unsigned int reserved2:6;  /* reserved */
	  } cesr;
	  unsigned int word;           /* to initialize in one assignment */
	} cesr;
	
	/* set CESR */
	cesr.word = 0;
	cesr.cesr.es0 = sel0;
	cesr.cesr.es1 = sel1;
	cesr.cesr.cc0 = cesr.cesr.cc1 = cpl;
	
	/* zero both counters. */
	clear = 0;
	lseek(fd, PENTIUM_MSR_CNT0, SEEK_SET);
	write(fd, &clear, 8);
	lseek(fd, PENTIUM_MSR_CNT1, SEEK_SET);
	write(fd, &clear, 8);
	
	/* Write out CESR. */
	lseek(fd, PENTIUM_MSR_CTLEVSEL, SEEK_SET);
	write(fd, &cesr, 8);
      }


    else if (event_table == event_table_pentium2)
      {
	union
	{
	  struct pentium2_cesr
	  {
	    unsigned int event_select:8;  /* event select */
	    unsigned int unit_mask:8;     /* further qualifies event (MESI) */
	    unsigned int usr:1;           /* count in user mode (cpl=1,2,3) */
	    unsigned int os:1;            /* count in os mode (cpl=0) */
	    unsigned int e:1;             /* edge detect */
	    unsigned int pc:1;            /* pin control */
	    unsigned int aint:1;          /* local APIC interrupt enable on
					     overflow */
	    unsigned int reserved:1;      /* reserved */
	    unsigned int en:1;            /* enable counters (sel0 only) */
	    unsigned int inv:1;           /* invert counter mask */
	    unsigned int counter_mask:8;  /* if!0, compare with events */
	  } cesr;
	  unsigned long long lword;    /* to initialize in one assignment */
	} cesr0, cesr1;
	
	    
	/* set all bits to 0 */
	cesr0.lword = cesr1.lword = 0;
	/* events */
	cesr0.cesr.event_select = sel0;
	cesr1.cesr.event_select = sel1;
	/* count events for all MESI states (applies only to some events) */
	if((sel0 == 0x29) || (sel0 == 0x2a) || (sel0 == 0x2e))
	  cesr0.cesr.unit_mask = 0xf;
	if((sel0 == 0x29) || (sel0 == 0x2a) || (sel0 == 0x2e))
	  cesr1.cesr.unit_mask = 0xf;
	/* additional fields */
	cesr0.cesr.usr = cesr1.cesr.usr = (cpl & 0x2) >> 1;
	cesr0.cesr.os = cesr1.cesr.os = cpl & 0x1;
	cesr0.cesr.en = cesr1.cesr.en = 1;
	/* all other fields get 0 by the above initialization */
	
	/* zero both counters. */
	clear = 0;
	lseek(fd, PENTIUM2_MSR_PERFCTR0, SEEK_SET);
	write(fd, &clear, 8);
	lseek(fd, PENTIUM2_MSR_PERFCTR1, SEEK_SET);
	write(fd, &clear, 8);
	
	/* Write out CESR0. */
	lseek(fd, PENTIUM2_MSR_EVNTSEL0, SEEK_SET);
	write(fd, &cesr0, 8);
	
	/* Write out CESR1. */
	lseek(fd, PENTIUM2_MSR_EVNTSEL1, SEEK_SET);
	write(fd, &cesr1, 8);
      }

    /* get actual clock value (to measure cycles) */
    rdtsc(&clock_value);

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)

  /* not implemented yet */


  /*--------------------------------------------------------------------------*/

#else
#error "unsupported system"
#endif
}


/*                                                                            */
/*============================================================================*/
/* read out counters.                                                         */


static int
hw_counter_read(
		  PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER] /* O: counter values */
)

{
  /*--------------------------------------------------------------------------*/
  /* system specific actions */

#if defined(ALPHA_CRAY)

  {
    long read_cycles;


    /* read out performance counters */
    if(ev5_perfctr_read(&counter_values[0], &counter_values[1],
			&counter_values[2]) != 0)
      return PCL_FAILURE;

    /* special case: cycle count */
    if(IS_USED(counter_used_mask, 3))
      {
	read_cycles = ev5_read_cycle_counter();
	if(read_cycles < cycle_counter_value)
	  /* overflow condition */
	  counter_values[3] = read_cycles + ((1l << 32) - cycle_counter_value);
	else
	  counter_values[3] = read_cycles - cycle_counter_value;
      }
  }


  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

  {
    int rc;
    long cycles, read_cycles;
    struct pfcntrs_ev5 cntrs_ev5[NCPUS];
    struct pfcntrs_ev5 *pfcntrs_ev5;
    union pmctrs_ev5 pmctrs_ev5;
#if defined(OS_HAS_SUPPORT_FOR_21264)
    /* the definition of struct pfcntrs in <sys/pfcntr.h> for the EV6 
       (which should be the same as for the EV4 as claimed by the docs)
       doesn't match the way data is stored by the pfm-driver. The definition
       in the include file uses the sequence <cntr0,cntr1,cycle> while the
       driver uses the following definition (which we have to use, too):
       */
    struct pfcntrs_ev6
      {
	ulong pr_cycle;    /* process cycle counter */
	ulong pf_cntr0;    /* driver's counter 1 */
	ulong pf_cntr1;    /* driver's counter 2 */
    };
    struct pfcntrs_ev6 cntrs_ev6[NCPUS];
    struct pfcntrs_ev6 *pfcntrs_ev6;
    union pmctrs_ev6 pmctrs_ev6;
#endif


    /* get actual cycle count from processor cycle counter */
    read_cycles = read_cycle_counter();
    if(read_cycles < cycle_counter_value)
      /* overflow condition */
      cycles = read_cycles + ((1l << 32) - cycle_counter_value);
    else
      cycles = read_cycles - cycle_counter_value;


    /* read out performance counters through driver interface */
    if(event_table == event_table_21164)
      {
	/* 21164 */
	pfcntrs_ev5 = &cntrs_ev5[cpu_num];

	/* get cpu counter values */
	pmctrs_ev5.pmctrs_ev5_long = 0;
	rc = ioctl(fd, PCNT5GETCNTRS, &pmctrs_ev5.pmctrs_ev5_long);
	if (rc < 0)
	  return PCL_FAILURE;

	/* get driver's interrupt values */
	rc = ioctl(fd, PCNT5GETCNT, &pfcntrs_ev5);
	if (rc < 0)
	  return PCL_FAILURE;

	/* extract counter values
	   (interrupts * timer_frequency + counter_values)
	   */
	counter_values[0] = (pfcntrs_ev5->pf_cntr0 << 16)
	  + pmctrs_ev5.pmctrs_ev5_cntr0;
	counter_values[1] = (pfcntrs_ev5->pf_cntr1 << 16)
	  + pmctrs_ev5.pmctrs_ev5_cntr1;
	counter_values[2] = (pfcntrs_ev5->pf_cntr2 << 14)
	  + pmctrs_ev5.pmctrs_ev5_cntr2;

	/* special case: elapsed cycles */
	counter_values[3] = cycles;
      }

#if defined(OS_HAS_SUPPORT_FOR_21264)
    else
      {
	/* 21264 */
	/* get cpu counter values */
	pmctrs_ev6.pmctrs_ev6_long = 0;
	rc = ioctl(fd, PCNT6GETCNTRS, &pmctrs_ev6.pmctrs_ev6_long);
	if (rc < 0)
	  return PCL_FAILURE;

	/* get driver's interrupt values */
	pfcntrs_ev6 = &cntrs_ev6[cpu_num];
	rc = ioctl(fd, PCNT6GETCNT, &pfcntrs_ev6);
	if (rc < 0)
	  return PCL_FAILURE;

	/* extract counter values
	   (interrupts * timer_frequency + counter_values) */
	/* timer frequency is fixed to 2^20 (1048576) */
	counter_values[0] = (pfcntrs_ev6->pf_cntr0 << 20)
	  + pmctrs_ev6.pmctrs_ev6_cntr0;
	counter_values[1] = (pfcntrs_ev6->pf_cntr1 << 20)
	  + pmctrs_ev6.pmctrs_ev6_cntr1;

	/* special case: elapsed cycles */
	counter_values[2] = cycles;
       }
#endif
  }


  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)
  {
    int sel0, sel1;
    hwperf_cntr_t cnts;


    /* read hardware counters */
    if(ioctl(fd, PIOCGETEVCTRS, (void *)&cnts) != generation_number)
      return PCL_FAILURE;

    /* Extract selected events and store result values. */
    if(IS_USED(counter_used_mask, 0))
      {
	sel0 = results[0].command;
	counter_values[0] = cnts.hwp_evctr[sel0];
      }
    if(IS_USED(counter_used_mask, 1))
      {
	sel1 = results[1].command;
	counter_values[1] = cnts.hwp_evctr[sel1];
      }
  }


  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

  {
    unsigned long long val, clocks;


    /* finish all outstanding operations */
    cpu_sync();

    /* get actual cycle count */
    clocks = get_tick();
    clocks = clocks - clock_value;

    /* get PIC-values as a difference between the actual value and the value
       gotten when starting the counts. The lower 32 bits are counter0 results,
       the upper 32 bits are counter1 results.
    */
    val = get_pic();
    counter_values[0] = (val & 0xffffffff) - pic0_value;
    counter_values[1] =  (val >> 32) - pic1_value;

    /* special case: cycle count. */
    counter_values[2] = clocks;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  {
    union PM_RETN retn;


    /* Read the accumulated values and store for further operations. */
    retn.w = PMstatus(&sc);
    if (retn.w != 0)
      return PCL_FAILURE;

    /* store result values */
    counter_values[0] = eil.ei[0].curr;
    counter_values[1] = eil.ei[1].curr;
    counter_values[2] = eil.ei[2].curr;
    counter_values[3] = eil.ei[3].curr;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)

  {
    long long clocks;
    unsigned int tmp[2];


    /* get actual cycle count */
    rdtsc(&clocks);
    counter_values[2] = clocks - clock_value;

    /* read counter values */
    if (can_use_rdpmc)
      {
	/* Pentium MMX, Pentium Pro, Pentium II, Pentium III */
	counter_values[0] = rdpmc(0);
	counter_values[1] = rdpmc(1);
      }
    else
      {
	/* Pentium */
	lseek(fd, PENTIUM_MSR_CNT0, SEEK_SET);
	read(fd, tmp, 8);
	counter_values[0] = (((long long)(tmp[1] & 0xff)) << 32 | tmp[0]);
	lseek(fd, PENTIUM_MSR_CNT0, SEEK_SET);
	read(fd, tmp, 8);
	counter_values[1] = (((long long)(tmp[1] & 0xff)) << 32 | tmp[0]);
      }
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)
  /* not yet implemented */


  /*--------------------------------------------------------------------------*/

#error "unsupported system"


  /*--------------------------------------------------------------------------*/

#else
#error "unsupported system"
#endif


  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* stop counters.                                                             */


static int
hw_counter_stop(void)

{
  /*--------------------------------------------------------------------------*/
  /* system specific actions */

#if defined(ALPHA_CRAY)

    {
      long ctr0, ctr1, ctr2;

#if defined(CHECK_MIGRATION)
      /* Check that we are still on the same processor. Eventually, the process
	 was migrated to another processor. In that case we have wrong numbers.
	 */
      if(processor_number != PCL_processor_number())
	return PCL_FAILURE;
#endif

      /* read out and stop performance counters */
      if(ev5_perfctr_stopread(&ctr0, &ctr1, &ctr2) != 0)
	return PCL_FAILURE;
    }

  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

    {
      int rc;
      union pmctrs_ev5 pm_ev5;
#if defined(OS_HAS_SUPPORT_FOR_21264)
      union pmctrs_ev6 pm_ev6;
#endif

    if(event_table == event_table_21164)
      {
	/* 21164 */
	/* disable counters*/
	pm_ev5.pmctrs_ev5_cpu = cpu_num;
	pm_ev5.pmctrs_ev5_select = counter_select;
	rc = ioctl(fd, PCNT5DISABLE, &pm_ev5.pmctrs_ev5_long);
	if (rc < 0)
	  return PCL_FAILURE;
      }

#if defined(OS_HAS_SUPPORT_FOR_21264)
    else
      {
	/* 21264 */
	/* disable counters*/
	pm_ev6.pmctrs_ev6_cpu = cpu_num;
	pm_ev6.pmctrs_ev6_select = counter_select;
	rc = ioctl(fd, PCNT6DISABLE, &pm_ev6.pmctrs_ev6_long);
	if (rc < 0)
	  return PCL_FAILURE;
      }
#endif
    }



  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)
    {
      if(ioctl(fd, PIOCRELEVCTRS) < 0)
	return PCL_FAILURE;
    }


  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

    /* ??? what happens here? */

  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  {
    union PM_RETN retn;

    /* Terminate counting */
    /* Set revision field */
    pc.revision = PM_PROCESS_CONTROL_REV;
    pc.command = PM_COUNT_TERM;
    PMprocess(&pc);
  }

  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)
    {
#if 0
      /* ???  Here, the msr device should be closed. But we don't do it to
      fasten up subsequent PCLstarts. This is ok as we currently have no
      operating system support to allow performance registers saves/restores on
      context switches. If we at some time include this code, we must change
      other parts also (i.e. reopen fd) */
      if(close(fd) == -1)
	return PCL_FAILURE;
      fd = -1;
#endif
    }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)

  /* not yet implemented */
#error "NT unsupported system"


  /*--------------------------------------------------------------------------*/

#else
#error "unsupported system"
#endif


    /* everything ok */
    return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Reads out hardware counters and stops counting.                            */


static int
hw_counter_read_and_stop(
	       PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER] /* O: counter values */
)

{
  int res;


  /* read out the current values */
  if((res = hw_counter_read(counter_values)) != PCL_SUCCESS)
    return res;

  /* stop counters */
  return hw_counter_stop();
}


/*                                                                            */
/*============================================================================*/
/* interrupt routine to avoid overflow.                                       */
/* 32-bit performance counters only                                           */


#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)

static void
interrupt_handling(
		    int sig
)

{
  int i, nlevel, res;
  PCL_CNT_TYPE counter_values[PCL_MAX_EVENT_PER_CALL];


  /* stop counting and read values */
  if (hw_counter_read_and_stop(counter_values) != PCL_SUCCESS)
    fprintf(stderr, "PCL: problem in interrupt routine (1)\n");

  /* add partial results to current nesting level */
  DEFINE_USE(virtual_counters_used_mask, nesting_level);
  for (i = 0; i < PCL_MAX_COUNTER; ++i)
    virtual_counters[nesting_level][i] += counter_values[i];

  /* restart counting and read out current values as base values
     for all nesting level */
  if ((hw_counter_start() != PCL_SUCCESS)
      || ((res = hw_counter_read(start_values[nesting_level])) != PCL_SUCCESS))
    fprintf(stderr, "PCL: problem in interrupt routine (2)\n");

  /* all other nesting levels are reset to the new starting values.
     The final results for each level get updated by the upper level values.
   */
  for(nlevel = nesting_level - 1; nlevel >= 0; --nlevel)
    for(i = 0; i < PCL_MAX_COUNTER; ++i)
      start_values[nlevel][i] = start_values[nesting_level][i];
}

#endif


/*                                                                            */
/*============================================================================*/
/* Try to match the event requests with the available counters.               */
/* First, try to match composed events as these are fixed to certain counters.*/
/* Then, try to match integral events to left counters.                       */
/* Not all possible mappings are successful as we have a simple strategy      */
/* might return "not possible" even if there is some combination.             */
/* Should be done better in the future.                                       */


static int
map_low_level_counters(
			int *counter_list,	/* I: events to be started */
			int ncounter,   /* I: number of events */
			result_descr_t results[PCL_MAX_COUNTER], /* O: cntr descr */
			bit_mask_t *counter_used_mask /* O: bit set to 1 if counter is used */
)

{
  int i, j, ecount, found;
  event_descr_t *event;
  result_descr_t *this_result;
  unsigned int tmp_mask;


  ecount = 0;

  /* generate bit mask what hardware counters are used */
  tmp_mask = 0;


  /*--------------------------------------------------------------------------*/
  /* handle each composed request */

  for (i = 0; i < ncounter; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &event_table[counter_list[i]];

      /* check if event is supported */
      if ((event->available == PCL_AVAIL) && (event->composed == COMPOSED))
	{
	  /* check for low level counters which match event */
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if (event->data.composed_data.command[j] != -1)
		{
		  if (!IS_USED(tmp_mask, j))
		    {
		      this_result = &results[j];

		      /* counter found */
		      /* this counter will be  used */
		      DEFINE_USE(tmp_mask, j);

		      /* command to activate cunter */
		      this_result->command = event->data.composed_data.command[j];

		      /* counter value is used for result i */
		      this_result->result_index = i;

		      /* combine operation */
		      this_result->operation =
			event->data.composed_data.operation[j];

		      /* store to what event type it belongs */
		      this_result->event = counter_list[i];
		    }

		  else
		    /* counter already used */
		    return PCL_NOT_SUPPORTED;
		}		       /* if(event->data.composed... */
	    }			       /* for(j = 0; j < PCL_MAX_COUNTER; ++j) */

	  /* one more event done */
	  ++ecount;
	}
    }



  /*--------------------------------------------------------------------------*/
  /* handle each integral request                                             */
  /* this is done in two steps: first map events that can only run on one     */
  /* specific counter only. In the second step map the rest of the events     */
  /* might run on more than one counter.                                      */

  /* step 1 */
  for (i = 0; i < ncounter; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &event_table[counter_list[i]];

      /* check if event is supported and not composed */
      if ((event->available == PCL_AVAIL) && (event->composed == 0))
	{
	  /* check if the event must run on one specific counter (and this
	     counter is still available) */
	  found = 0;
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if ((event->data.command[j] != -1) &&
		  (!IS_USED(tmp_mask, j)))
		++found;
	    }
	  if(found != 1)
	    continue;

	  /* check for low level counter which matches event and  which is
	     unused */
	  found = 0;
	  for (j = 0; (j < PCL_MAX_COUNTER) && !found; ++j)
	    {
	      if ((event->data.command[j] != -1) &&
		  (!IS_USED(tmp_mask, j)))
		{
		  /* this counter will be  used */
		  DEFINE_USE(tmp_mask, j);

		  /* counter found */
		  this_result = &results[j];

		  /* command to activate counter */
		  this_result->command = event->data.command[j];

		  /* counter value is used for result i */
		  this_result->result_index = i;

		  /* integral value (add value)  */
		  this_result->operation = OP_ADD;

		  /* store to what event type it belongs */
		  this_result->event = counter_list[i];

		  /* one more event done */
		  ++ecount;

		  /* leave loop */
		  found = 1;
		}
	    }
	}
    }


  /* step 2 */
  for (i = 0; i < ncounter; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &event_table[counter_list[i]];

      /* check if event is supported and not composed */
      if ((event->available == PCL_AVAIL) && (event->composed == 0))
	{
	  /* check if the event can run on more than one counter */
	  found = 0;
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if ((event->data.command[j] != -1))
		++found;
	    }
	  if(found < 2)
	    continue;

	  /* check for low level counter which matches event and  which is
	     unused */
	  found = 0;
	  for (j = 0; (j < PCL_MAX_COUNTER) && !found; ++j)
	    {
	      if ((event->data.command[j] != -1) &&
		  (!IS_USED(tmp_mask, j)))
		{
		  /* this counter will be  used */
		  DEFINE_USE(tmp_mask, j);

		  /* counter found */
		  this_result = &results[j];

		  /* command to activate counter */
		  this_result->command = event->data.command[j];

		  /* counter value is used for result i */
		  this_result->result_index = i;

		  /* integral value (add value)  */
		  this_result->operation = OP_ADD;

		  /* store to what event type it belongs */
		  this_result->event = counter_list[i];

		  /* one more event done */
		  ++ecount;

		  /* leave loop */
		  found = 1;
		}
	    }
	}
    }


  /*--------------------------------------------------------------------------*/
  /* check if we got the whole mapping */

  if (ecount != ncounter)
    /* at least one event is not supported */
    return PCL_NOT_SUPPORTED;

  /* store what counter are used */
  *counter_used_mask = tmp_mask;

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Based on the result values of the low-level counters, calculate results.   */


static void
generate_results(
		  PCL_CNT_TYPE * results_list,             /* O: counter values */
		  int ncounter,                            /* I: number of counter values */
		  result_descr_t results[PCL_MAX_COUNTER]  /* I: result descr */
		  )

{
  int j, ind;
  unsigned int op;
  result_descr_t *this_result;
  double *dp, *dp1, d;


  /* initialize result array */
  for(j = 0; j < ncounter; ++j)
    results_list[j] = 0;

  /* loop over all counters */
  for (j = 0; j < PCL_MAX_COUNTER; ++j)
    {
      if(IS_USED(counter_used_mask, j))
	{
	  this_result = &results[j];

	  /* calculate index in result array where to put result value */
	  ind = this_result->result_index;

	  /* get operation specifier */
	  op = this_result->operation;

	  if(op & OP_FP_TEST_MASK)
	    {
	      /* change 1st operand type from int to double? */
	      if (op & OP_FP_1)
		{
		  dp = (double *)&results_list[ind];
		  d = (double)results_list[ind];
		  *dp = d;
		}
	      
	      /* change 2nd operand type from int to double? */
	      if (op & OP_FP_2)
		{
		  dp = (double *)&this_result->result_value;
		  d = (double)this_result->result_value;
		  *dp = d;
		}

	      /* mask out conversion operations */
	      op &= OP_FP_MASK;
	    }


	  /* apply operation */
	  switch(op)
	    {
	    case OP_ADD:
	      /* add to result value the value of the j-th low-level counter */
	      results_list[ind] += this_result->result_value;
	      break;

	    case OP_SUB:
	      /* subtract from result value the value of the
		 j-th low-level counter
		 */
	      results_list[ind] -= this_result->result_value;
	      break;

	    case OP_FP_STORE:
	      /* store value in j-th low-level counter (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      *dp = *dp1;
	      break;

	    case OP_FP_ADD:
	      /* add value in j-th low-level counter to value
		 (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      *dp += *dp1;
	      break;

	    case OP_FP_MUL:
	      /* multiply value by the value of the j-th low-level
		 counter (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      *dp *= *dp1;
	    break;

	    case OP_FP_DIV:
	      /* divide value by the value of the j-th low-level
		 counter (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      if(*dp1 == 0.0)
		*dp1 = 1.0;
	      *dp /= *dp1;
	      break;

	    case OP_FP_DIV_MHZ:
	    {
	      /* divide value by value in the j-th low-level counter
		 and multiply by MHz-rate (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      if(*dp1 <= 0.0)
		*dp1 = 1.0;
	      *dp = *dp * (double)mhz_rate / *dp1;
	    }

	    case OP_FP_REZ:
	      /* add value by the repricoc value of the
		 j-th low-level counter (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      if(*dp1 <= 0.0)
		*dp1 = 1.0;
	      *dp += 1.0 / *dp1;
	      break;

	    case OP_FP_REZ_MHZ:
	      /* add value by the repricoc value of the
		 j-th low-level counter and multiply by MHz-rate
		 (floating point) */
	      dp = (double *)&results_list[ind];
	      dp1 = (double *)&this_result->result_value;
	      if(*dp1 <= 0.0)
		*dp1 = 1.0;
	      *dp += (double)mhz_rate / *dp1;
	      break;

	    default:
	      fprintf(stderr, "PCL: internal error (illegal operation)\n");
	    }
	}
    }
}


/*                                                                            */
/*============================================================================*/
/* check input arguments to PCLquery and PCLstart                             */


static int
check_and_init(
	       int *counter_list,     /* I: event counters to be started */
	       int ncounter,	       /* I: number of event counters */
	       unsigned int flags     /* I: mode flags (PCL_MODE_xxx) */
	       )

{
#if defined(CHECK_ARGUMENTS)
  int i;
#endif


  /*--------------------------------------------------------------------------*/
  /* initialization (needed on some CPUs at this place to get the right 
     event table) */

#if defined(ALPHA_CRAY) && defined(CHECK_MIGRATION)
  if (!initialized)
    {
      /* do this only once*/
      initialized = 1;

      /* get processor number we are running on */
      processor_number = PCL_processor_number();
    }

#elif defined(ALPHA_DEC)
  if (!initialized)
    {
      int res;

      /* do this only once*/
      initialized = 1;

      if((res = initialize_21x64()) != PCL_SUCCESS)
	return res;
    }

#elif defined(MIPS)
  if (!initialized)
    {
      int res;

      /* do this only once*/
      initialized = 1;

      if((res = initialize_r1x000()) != PCL_SUCCESS)
	return res;
    }

#elif defined(PENTIUM_LINUX)
  if (!initialized)
    {
      int res;

      /* do this only once*/
      initialized = 1;

      if((res = initialize_pentium()) != PCL_SUCCESS)
	return res;
    }
#endif


  /* determine MHz-rate of processor */
  if(mhz_rate == -1)
    mhz_rate = PCL_determine_mhz_rate();


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

#if defined(CHECK_ARGUMENTS)

  /* check that number of requested events is smaller than maximum allowed
     number */
  if (ncounter >= PCL_MAX_EVENT_PER_CALL)
    return PCL_TOO_MANY_EVENTS;

  /* check that event identifiers are in allowed range */
  for (i = 0; i < ncounter; ++i)
    if ((counter_list[i] < 0) || (counter_list[i] >= PCL_MAX_EVENT))
      return PCL_ILL_EVENT;

  /* check correct mode */
  if((flags != PCL_MODE_USER)
     && (flags != PCL_MODE_SYSTEM)
     && (flags != PCL_MODE_USER_SYSTEM))
    return PCL_MODE_NOT_SUPPORTED;

  /* check if count mode is supported */
  if((flags & allowed_count_mode) != flags)
    return PCL_MODE_NOT_SUPPORTED;
#endif


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* query for functionality.                                                   */
/* returns PCL_SUCCESS if requested event counting is possible                */


/* Fortran version */
int
fname(pclquery, PCLQUERY)(
	   int *counter_list,	       /* I: requested event counters */
	   int *ncounter,	       /* I: # of requested event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  return PCLquery(counter_list, *ncounter, *flags);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLquery(
		  JNIEnv *env,
		  jobject obj,
		  jintArray counter_list, /* I: requested event counters */
		  jint ncounter,          /* I: # of requested event counters */
		  jint flags              /* I: mode flags (PCL_MODE_xxx) */
		  )

{
  jsize len;
  jint *body;
  int res;


  /* get array descriptor */
  len = (*env)->GetArrayLength(env, counter_list);
  body = (*env)->GetIntArrayElements(env, counter_list, 0);

  /* call PCL function */
  res = PCLquery((int *)body, ncounter, flags);

  /* release copied array */
  (*env)->ReleaseIntArrayElements(env, counter_list, body, 0);

  return res;
}
#endif


/* C, C++ version */
int
PCLquery(
	   int *counter_list,	       /* I: requested event counters */
	   int ncounter,	       /* I: # of requested event counters */
	   unsigned int flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  int res;
  result_descr_t tmp_results[PCL_MAX_COUNTER];
  bit_mask_t tmp_mask;


  /* check call arguments */
  if((res = check_and_init(counter_list, ncounter, flags)) != PCL_SUCCESS)
    return res;

  /* Try to get a mapping of events to counters in a temporary counter list.
     If this can be done, the query is successful.
  */
  return map_low_level_counters(counter_list, ncounter, tmp_results, &tmp_mask);
}


/*                                                                            */
/*============================================================================*/
/* start performance counting                                                 */
/* returns PCL_SUCCESS if everything went ok                                  */
/* If this is a nested call we do the following:                              */
/*    - increment the nesting level                                           */
/*    - check that same events are requested than in previous level           */
/*    - stops previous couting and store result values in <partial_results>   */


/* Fortran version */
int
fname(pclstart, PCLSTART)(
	   int *counter_list,	       /* I: event counters to be started */
	   int *ncounter,	       /* I: number of event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  return PCLstart(counter_list, *ncounter, *flags);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLstart(
		  JNIEnv *env,
		  jobject obj,
		  jintArray counter_list, /* I: requested event counters */
		  jint ncounter,          /* I: # of requested event counters */
		  jint flags              /* I: mode flags (PCL_MODE_xxx) */
		  )

{
  jsize len;
  jint *body;
  int res;


  /* get array descriptor */
  len = (*env)->GetArrayLength(env, counter_list);
  body = (*env)->GetIntArrayElements(env, counter_list, 0);

  /* call PCL function */
  res = PCLstart((int *)body, ncounter, flags);

  /* release copied array */
  (*env)->ReleaseIntArrayElements(env, counter_list, body, 0);

  return res;
}
#endif


/* C, C++ version */
int
PCLstart(
	   int *counter_list,	       /* I: event counters to be started */
	   int ncounter,	       /* I: number of event counters */
	   unsigned int flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  int i, res;


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  if((res = check_and_init(counter_list, ncounter, flags)) != PCL_SUCCESS)
    return res;

#if defined(CHECK_ARGUMENTS)
  /* check for incrementing nesting */
  if (nesting_level + 1 >= PCL_MAX_NESTING_LEVEL)
    return PCL_TOO_MANY_NESTINGS;
#endif


  /*--------------------------------------------------------------------------*/
  /* initialize nesting level */

  /* increment nesting level */
  ++nesting_level;

  /* so far, the virtual counters for this nesting level are not used */
  RESET_USE(virtual_counters_used_mask, nesting_level);

  /* set virtual counters and start values for this nesting level to 0 */
  for(i = 0; i < PCL_MAX_COUNTER; ++i)
    {
      virtual_counters[nesting_level][i] = 0;
      start_values[nesting_level][i] = 0;
    }


  /*--------------------------------------------------------------------------*/
  /* handle previous calls */

  if (nesting_level == 0)
    {
      /* outermost nesting level */
      /* store request for this call */
      event_count = ncounter;
      for (i = 0; i < ncounter; ++i)
	events[i] = counter_list[i];

      /* store count mode */
      count_mode = flags;

      /* try to get a mapping of events to counters */
      if((res = map_low_level_counters(counter_list, ncounter, results,
				       &counter_used_mask)) != PCL_SUCCESS)
	return res;

      fp_result_mask = 0;
      for(i = 0; i < ncounter; ++i)
	if(counter_list[i] >= PCL_MFLOPS)
	  DEFINE_USE(fp_result_mask, i);

      /* start the counters */
      if((res = hw_counter_start()) != PCL_SUCCESS)
	return res;
    }

  else
    {
      /* inner nesting level */

      /* check that the same events are requested as in lower nesting levels */
#if defined(CHECK_ARGUMENTS)
      /* check number of requests */
      if (ncounter != event_count)
	return PCL_ILL_NESTING;

      /* check that event requests of this call are the same than that of
         previous call */
      for (i = 0; i < ncounter; ++i)
	if (counter_list[i] != events[i])
	  return PCL_ILL_NESTING;

      /* check correct mode */
      if(flags != count_mode)
	return PCL_MODE_NOT_SUPPORTED;
#endif

      /* read out current values as base values for this nesting level */
      if((res = hw_counter_read(start_values[nesting_level]))
	 != PCL_SUCCESS)
      return res;
    }


  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* read out counter values and process values                                 */


static int
do_read_or_stop(
		PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
		PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
		int ncounter,                     /* I: number of event counters */
		int do_restart                    /* I: 1=restart counting, 0=stop */
		)


{
  int i, res;
  PCL_CNT_TYPE counter_values[PCL_MAX_COUNTER];


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

#if defined(CHECK_ARGUMENTS)
  /* check call level and requests for this call with corresponding PCLstart */
  if ((nesting_level < 0) || (event_count != ncounter))
    return PCL_ILL_NESTING;
#endif


  /*--------------------------------------------------------------------------*/
  /* read values */

  if(do_restart)
    {
      /* read out values */
      if ((res = hw_counter_read(counter_values)) != PCL_SUCCESS)
	return res;
    }
  else
    {
      /* read out values and stop counters */
      if ((res = hw_counter_read_and_stop(counter_values) != PCL_SUCCESS))
	return res;
    }



  /* subtract start values gotten with PCLstart and add partial results */
  if(IS_USED(virtual_counters_used_mask, nesting_level))
    {
      for (i = 0; i < PCL_MAX_COUNTER; ++i)
	/* if(IS_USED(counter_used_mask, i)) */
	  results[i].result_value =
	    (counter_values[i] - start_values[nesting_level][i])
	    + virtual_counters[nesting_level][i];
    }
  else
    {
      for (i = 0; i < PCL_MAX_COUNTER; ++i)
	/* if(IS_USED(counter_used_mask, i))*/
	  results[i].result_value =
	    counter_values[i] - start_values[nesting_level][i];
    }

  /* generate results from performance counter values */
  generate_results(i_result_list, ncounter, results);

  /* transfer floating point results to fp_result_list */
  if(fp_result_mask)
    {
      for(i = 0; i < event_count; ++i)
	if(IS_USED(fp_result_mask, i))
	  fp_result_list[i] = *(PCL_FP_CNT_TYPE *)&i_result_list[i];
    }

  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* reads performance counters                                                 */
/* result_list must be a pointer to a sufficiently large buffer.              */
/* returns PCL_SUCCESS if everything went ok                                  */


/* Fortran version */
int
fname(pclread, PCLREAD)(
	  PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
	  PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	  int *ncounter		            /* I: number of event counters */
)

{
  return PCLread(i_result_list, fp_result_list, *ncounter);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLread(
		  JNIEnv *env,
		  jobject obj,
		  jlongArray i_result_list,  /* O: integer counter values */
		  jdoubleArray fp_result_list, /* O: FP counter values */
		  jint ncounter              /* I: number of event counters */
		  )

{
  jsize i_len, fp_len;
  jlong *i_body;
  jdouble *fp_body;
  int res;


  /* get array descriptor */
  i_len = (*env)->GetArrayLength(env, i_result_list);
  fp_len = (*env)->GetArrayLength(env, fp_result_list);
  i_body = (*env)->GetLongArrayElements(env, i_result_list, 0);
  fp_body = (*env)->GetDoubleArrayElements(env, fp_result_list, 0);

  /* call PCL function */
  res = PCLread(i_body, fp_body, ncounter);

  /* release copied array */
  (*env)->ReleaseLongArrayElements(env, i_result_list, i_body, 0);
  (*env)->ReleaseDoubleArrayElements(env, fp_result_list, fp_body, 0);

  return res;
}
#endif


/* C, C++ version */
int
PCLread(
	  PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
	  PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	  int ncounter		            /* I: number of event counters */
)

{
  /* get results */
  return do_read_or_stop(i_result_list, fp_result_list, ncounter, 1);
}


/*                                                                            */
/*============================================================================*/
/* stops performance counting and reads counters values.                      */
/* result_list must be a pointer to a sufficiently large buffer.              */
/* returns PCL_SUCCESS if everything went ok                                  */


/* Fortran version */
int
fname(pclstop, PCLSTOP)(
	  PCL_CNT_TYPE * i_result_list,      /* O: integer counter values */
	  PCL_FP_CNT_TYPE * fp_result_list,  /* O: FP counter values */
	  int *ncounter		             /* I: number of event counters */
)

{
  return PCLstop(i_result_list, fp_result_list, *ncounter);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jint JNICALL
Java_PCL_cPCLstop(
		  JNIEnv *env,
		  jobject obj,
		  jlongArray i_result_list,     /* O: integer counter values */
		  jdoubleArray fp_result_list,  /* O: FP counter values */
		  jint ncounter                 /* I: number of event counters */
		  )

{
  jsize i_len, fp_len;
  jlong *i_body;
  jdouble *fp_body;
  int res;


  /* get array descriptor */
  i_len = (*env)->GetArrayLength(env, i_result_list);
  fp_len = (*env)->GetArrayLength(env, fp_result_list);
  i_body = (*env)->GetLongArrayElements(env, i_result_list, 0);
  fp_body = (*env)->GetDoubleArrayElements(env, fp_result_list, 0);

  /* call PCL function */
  res = PCLstop(i_body, fp_body, ncounter);

  /* release copied array */
  (*env)->ReleaseLongArrayElements(env, i_result_list, i_body, 0);
  (*env)->ReleaseDoubleArrayElements(env, fp_result_list, fp_body, 0);
  return res;
}
#endif

/* C, C++ version */
int
PCLstop(
	  PCL_CNT_TYPE * i_result_list,      /* O: integer counter values */
	  PCL_FP_CNT_TYPE * fp_result_list,  /* O: FP counter values */
	  int ncounter		             /* I: number of event counters */
)

{
  int i, res;
#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)
  /* 32-bit performance counters */
  struct itimerval itimer;
#endif


  /*--------------------------------------------------------------------------*/

  /* check arguments and read out values */
  if((res = do_read_or_stop(i_result_list, fp_result_list, ncounter,
			    nesting_level > 0)) != PCL_SUCCESS)
    return res;

  if (nesting_level > 0)
    {
      /* add virtual counter values to previous nesting level */
      if(IS_USED(virtual_counters_used_mask, nesting_level))
	for(i = 0; i < PCL_MAX_COUNTER; ++i)
	  virtual_counters[nesting_level-1][i] += virtual_counters[nesting_level][i];
    }

  /* decrement nesting level*/
  --nesting_level;


  /*--------------------------------------------------------------------------*/
  /* stop timer interrupts (overflow avoidance) */

#if defined(COUNTER_SIZE_32) && defined(COUNTER32_ARE_NOT_ENOUGH)
  /* 32-bit performance counters */
  if (nesting_level == -1)
    {
      /* set up signal handler */
      if (signal(ITIMER_SIGNAL, old_signal_handler) == SIG_ERR)
	return PCL_FAILURE;

      /* cancel interval timer */
      itimer.it_interval.tv_sec = 0;
      itimer.it_interval.tv_usec = 0;
      itimer.it_value.tv_sec = 0;
      itimer.it_value.tv_usec = 0;
      setitimer(ITIMER_MODE, &itimer, NULL);

      old_signal_handler = NULL;
    }
#endif


  /*--------------------------------------------------------------------------*/
  /* leave funtion */

  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* returns name (character string) of event.           .                      */


/* Fortran version */
const char *
fname(pcleventname, PCLEVENTNAME)(
		  int *event		       /* I: event to get name for */
                 )

{
    /* ??? not yet done
     Fortran character passing is one of the few unsolved problems
     in computer science. We don't solve it for the moment.
    */
    return PCLeventname(*event);
}


/* Java version */
#if defined(JAVA_INTERFACE)
JNIEXPORT jstring JNICALL
Java_PCL_cPCLeventname(
		       JNIEnv *env,
		       jobject obj,
		       jint event           /* I: event to get name for */
		       )

{
  const char *res;


  /* call PCL function */
  res = PCLeventname(event);

  return (*env)->NewStringUTF(env, res);
}
#endif


const char *
PCLeventname(
	     int event		       /* I: event to get name for */
)

{
  if ((event < 0) || (event >= PCL_MAX_EVENT))
    return (const char *)NULL;
  else
    return eventname[event];
}


/*                                                                            */
/*============================================================================*/
/* Return the value of an elapsed cycle counter if available, otherwise -1.   */
/* The functionality exists also possible with PCLstart/PCLstop but sometimes */
/* it is important to have a very fast function.                              */
/* Currently, this is not part of the official PCL interface.                 */


PCL_CNT_TYPE
PCLelapsed_cycle_count(void)
{
#if defined(ALPHA_CRAY)
  return ev5_read_cycle_counter();

#elif defined(ALPHA_DEC)
  return read_cycle_counter();

#elif defined(SPARC)
  return get_tick();

#elif defined(MIPS)
  /*???*/
  return -1;

#elif defined(POWERPC)
  /*???*/
  return -1;

#elif defined(PENTIUM_LINUX)
  long long ticks;

  /* get actual clock value (to measure cycles) */
  rdtsc(&ticks);
  return ticks;
#endif
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/

