/*==============================================================================


   Purpose:    interface for Performance Counter Library (PCL)


   Author:     Heinz Ziegler, Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   05.07.1999


   DISCLAIMER

                  PCL - The Performance Counter Library


   Copyright (C) 1999, Forschungszentrum Juelich GmbH, Federal Republic of
   Germany. All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

     - Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

     - Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

     - Any publications that result from the use of this software shall
       reasonably refer to the Research Centre's development.

     - All advertising materials mentioning features or use of this software
       must display the following acknowledgement:

           This product includes software developed by Forschungszentrum
           Juelich GmbH, Federal Republic of Germany.

     - Forschungszentrum Juelich GmbH is not obligated to provide the user with
       any support, consulting, training or assistance of any kind with regard
       to the use, operation and performance of this software or to provide
       the user with any updates, revisions or new versions.


   THIS SOFTWARE IS PROVIDED BY FORSCHUNGSZENTRUM JUELICH GMBH "AS IS" AND ANY
   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   DISCLAIMED. IN NO EVENT SHALL FORSCHUNGSZENTRUM JUELICH GMBH BE LIABLE FOR
   ANY SPECIAL, DIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE ACCESS, USE OR PERFORMANCE OF THIS SOFTWARE.

==============================================================================*/


#if !defined(PCL_INCLUDED)
#define PCL_INCLUDED


/*                                                                            */
/*============================================================================*/
/* define exactly one of the following systems:                               */


/* you can do that either here or in the Makefile with -Dsystem_name */
/*#define ALPHA_CRAY */
/*#define ALPHA_DEC */
/*#define MIPS */
/*#define SPARC */
/*#define POWERPC */
/*#define PENTIUM_LINUX */
/*#define PENTIUM_NT */


/*                                                                            */
/*============================================================================*/
/* Include hardware/os dependend libraries                                    */


#include <stdio.h>
#include <stdlib.h>


/*                                                                            */
/*============================================================================*/
/* Generate Fortran-names out of C-names                                      */


#if defined(ALPHA_DEC) || defined(MIPS) || defined(SPARC) || defined(PENTIUM_LINUX)
/* all lowercase, underscore appended */
#define fname(n1, n2) n1##_
#elif defined(ALPHA_CRAY)
/* uppercase letters */
#define fname(n1, n2) n2
#else
/* all lowercase */
#define fname(n1, n2) n1
#endif


/*                                                                            */
/*============================================================================*/
/* some constants:                                                            */


/* maximum values */
#define PCL_MAX_EVENT          61      /* number of supported events */
#define PCL_MAX_EVENT_PER_CALL  4      /* max. number of events per call */
#define PCL_MAX_NESTING_LEVEL  16      /* max. number of nested calls to PCL */


/* Upper bound on the number of hardware counters. This might be useful for
   dimensioning (static) arrays.
 */
#define PCL_COUNTER_MAX 4


/* return codes */
#define PCL_SUCCESS             0      /* successful operation */
#define PCL_NOT_SUPPORTED      -1      /* failure: event not supported */
#define PCL_TOO_MANY_EVENTS    -2      /* failure: not all requests supported */
#define PCL_TOO_MANY_NESTINGS  -3      /* too many call nesting */
#define PCL_ILL_NESTING        -4      /* illegal call nesting */
#define PCL_ILL_EVENT          -5      /* illegal event identifier */
#define PCL_MODE_NOT_SUPPORTED -6      /* mode not supported */
#define PCL_FAILURE            -7      /* failure: something else */


/* result type for performance counters (64 bit integer) */
#if defined(ALPHA_CRAY)
#define PCL_CNT_TYPE long

#elif defined(ALPHA_DEC)
#define PCL_CNT_TYPE long

#elif defined(MIPS)
#define PCL_CNT_TYPE long long

#elif defined(SPARC)
#define PCL_CNT_TYPE long long

#elif defined(POWERPC)
#define PCL_CNT_TYPE long long

#elif defined(PENTIUM_LINUX)
#define PCL_CNT_TYPE long long

#elif defined(PENTIUM_NT)
#define PCL_CNT_TYPE long long

#else
#error "unsupported system"

#endif


/* result type for performance counters (64 bit floating point) */
#define PCL_FP_CNT_TYPE double



/*                                                                            */
/*============================================================================*/
/* mode flags:                                                                */


/* define what to count */
#define PCL_MODE_USER         1u   /* count for user mode only */
#define PCL_MODE_SYSTEM       2u   /* count for system mode only */
#define PCL_MODE_USER_SYSTEM  (PCL_MODE_USER|PCL_MODE_SYSTEM)  /* count both */


/*                                                                            */
/*============================================================================*/
/* supported events:                                                          */


/* Level-1-Cache */
#define PCL_L1CACHE_READ            0
#define PCL_L1CACHE_WRITE           1
#define PCL_L1CACHE_READWRITE       2
#define PCL_L1CACHE_HIT             3
#define PCL_L1CACHE_MISS            4

/* Level-1-Data-Cache */
#define PCL_L1DCACHE_READ           5
#define PCL_L1DCACHE_WRITE          6
#define PCL_L1DCACHE_READWRITE      7
#define PCL_L1DCACHE_HIT            8
#define PCL_L1DCACHE_MISS           9

/* Level-1-Instruction-Cache */
#define PCL_L1ICACHE_READ           10
#define PCL_L1ICACHE_WRITE          11
#define PCL_L1ICACHE_READWRITE      12
#define PCL_L1ICACHE_HIT            13
#define PCL_L1ICACHE_MISS           14

/*----------------------------------------------------------------------------*/

/* Level-2-Cache */
#define PCL_L2CACHE_READ            15
#define PCL_L2CACHE_WRITE           16
#define PCL_L2CACHE_READWRITE       17
#define PCL_L2CACHE_HIT             18
#define PCL_L2CACHE_MISS            19

/* Level-2-Data-Cache */
#define PCL_L2DCACHE_READ           20
#define PCL_L2DCACHE_WRITE          21
#define PCL_L2DCACHE_READWRITE      22
#define PCL_L2DCACHE_HIT            23
#define PCL_L2DCACHE_MISS           24

/* Level-2-Instruction-Cache */
#define PCL_L2ICACHE_READ           25
#define PCL_L2ICACHE_WRITE          26
#define PCL_L2ICACHE_READWRITE      27
#define PCL_L2ICACHE_HIT            28
#define PCL_L2ICACHE_MISS           29

/*----------------------------------------------------------------------------*/

/* TLB */
#define PCL_TLB_HIT                 30
#define PCL_TLB_MISS                31

/* Instruction-TLB */
#define PCL_ITLB_HIT                32
#define PCL_ITLB_MISS               33

/* Data-TLB */
#define PCL_DTLB_HIT                34
#define PCL_DTLB_MISS               35

/*----------------------------------------------------------------------------*/

/* Cycles */
#define PCL_CYCLES                  36
#define PCL_ELAPSED_CYCLES          37

/* Operations */
#define PCL_INTEGER_INSTR           38
#define PCL_FP_INSTR                39
#define PCL_LOAD_INSTR              40
#define PCL_STORE_INSTR             41
#define PCL_LOADSTORE_INSTR         42
#define PCL_INSTR                   43

/* Jump Instructions */
#define PCL_JUMP_SUCCESS            44
#define PCL_JUMP_UNSUCCESS          45
#define PCL_JUMP                    46

/* Atomic Instructions */
#define PCL_ATOMIC_SUCCESS          47
#define PCL_ATOMIC_UNSUCCESS        48
#define PCL_ATOMIC                  49

/*----------------------------------------------------------------------------*/

/* Stalls */
#define PCL_STALL_INTEGER           50
#define PCL_STALL_FP                51
#define PCL_STALL_JUMP              52
#define PCL_STALL_LOAD              53
#define PCL_STALL_STORE             54
#define PCL_STALL                   55

/*----------------------------------------------------------------------------*/
/* starting with PCL_MFLOPS, the results are 64-bit floating point numbers */

/* derived numbers */
#define PCL_MFLOPS                  56
#define PCL_IPC                     57
#define PCL_L1DCACHE_MISSRATE       58
#define PCL_L2DCACHE_MISSRATE       59
#define PCL_MEM_FP_RATIO            60


/*                                                                            */
/*============================================================================*/
/* C++:                                                                      */


#ifdef	__cplusplus
extern "C" {
#endif


/*                                                                            */
/*============================================================================*/
/* global variables:                                                          */


/* PCL version number */
extern const char *PCL_version_number;


/*                                                                            */
/*============================================================================*/
/* interface functions:                                                       */


/* query for functionality.
   returns PCL_SUCCESS if requested event counting is possible
 */
int PCLquery(
	       int *counter_list,      /* I: requested event counters */
	       int ncounter,	       /* I: number of requested event
				          counters */
	       unsigned int flags      /* I: mode flags (PCL_MODE_xxx) */
);

/* Fortran version */
int
fname(pclquery, PCLQUERY)(
	   int *counter_list,	       /* I: requested event counters */
	   int *ncounter,	       /* I: # of requested event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
);


/* start performance counting
   returns PCL_SUCCESS if everything went ok
 */
int PCLstart(
	       int *counter_list,      /* I: event counters to be started */
	       int ncounter,	       /* I: number of event counters */
	       unsigned int flags      /* I: flags (PCL_MODE_xxx) */
);

/* Fortran version */
int
fname(pclstart, PCLSTART)(
	   int *counter_list,	       /* I: event counters to be started */
	   int *ncounter,	       /* I: number of event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
);


/* reads performance counters.
   result_list must be a pointer to a sufficiently large buffer.
   returns PCL_SUCCESS if everything went ok
 */
int PCLread(
	      PCL_CNT_TYPE * i_result_list,     /* O: integer counter values */
	      PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	      int ncounter	                /* I: number of event counters */
);

/* Fortran version */
int
fname(pclread, PCLREAD)(
	  PCL_CNT_TYPE * i_results_list,    /* O: integer counter values */
	  PCL_FP_CNT_TYPE *fp_result_list,  /* O: FP counter values */
	  int *ncounter		            /* I: number of event counters */
);


/* stops performance counting and reads counters values.
   result_list must be a pointer to a sufficiently large buffer.
   returns PCL_SUCCESS if everything went ok
 */
int PCLstop(
	      PCL_CNT_TYPE * i_result_list,     /* O: counter values */
	      PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	      int ncounter	                /* I: number of event counters */
);

/* Fortran version */
int
fname(pclstop, PCLSTOP)(
	  PCL_CNT_TYPE * i_results_list,    /* O: counter values */
	  PCL_FP_CNT_TYPE * fp_result_list, /* O: FP counter values */
	  int *ncounter		            /* I: number of event counters */
);


/* get name (as character string) of an event. C-version only.
   returns string-pointer if everything went ok, else NULL
 */
const char *
PCLeventname(
	     int event                 /* I: event to get name for */
	     );


/*                                                                            */
/*============================================================================*/
/*                                                                            */

#ifdef	__cplusplus
}
#endif


/*                                                                            */
/*============================================================================*/
/* C++:                                                                      */


#endif				       /* !defined(PCL_INCLUDED) */


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
