/*==============================================================================


   Purpose:    test program for Performance Counter Library (PCL)


   Author:     Heinz Ziegler, Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   22.07.1999


==============================================================================*/

/* This is a program to test some of the functionality of the Performance
   Counter Library PCL. Execute "YOUR_BINARY_PATH/pcl_test.exe -help" to get a
   list of possible command arguments. Some of the tests execute a matrix
   addition and show results for that small kernel.
 */

#include "pcl.h"
#include "determine_mhz.h"

#include <string.h>
#include <sys/time.h>
#include <unistd.h>


/* dimension of test arrays */
#define N 200

/* number of iterations to get overhead */
#define NITER 300000

/* output format and type for results */
#define OFORMAT   "%13.0f"
#define OTYPE     double

/* error messages as strings */
static const char *(errornames[]) =
{
  "success",
	  "not supported",
	  "too many events",
	  "too many nestings",
	  "illegal nesting",
	  "illegal event",
	  "mode not supported",
	  "failure"
};


/* forward declaration of s small dummy function */
void dummy(unsigned int ijk);


/*                                                                            */
/*============================================================================*/
/* calculate the difference of two timeval numbers.                           */


double
timediff(struct timeval tv2, struct timeval tv1)
{
  return (double) (tv2.tv_sec - tv1.tv_sec)
	  + ((double) (tv2.tv_usec - tv1.tv_usec) * 1e-6);
}


/*                                                                            */
/*============================================================================*/
/* matrix addition as a simple test program.                                  */


static void
matadd(double a[N][N], double b[N][N], double c[N][N])

{
  int i, j;


  for (i = 0; i < N; ++i)
    for (j = 0; j < N; ++j)
      a[i][j] = b[i][j] + c[i][j];
}


/*                                                                            */
/*============================================================================*/
/* print either integer value or double value                                 */


static void
print_result(int event, PCL_CNT_TYPE i_result, PCL_FP_CNT_TYPE fp_result)
{
  if (event < PCL_MFLOPS)
    /* use %f format as we don't know what format we must use for 64-bit ints */
    printf("%13.0f", (double) i_result);
  else
    printf("%13f", fp_result);
}


/*                                                                            */
/*============================================================================*/
/* show usage information                                                     */


static void
usage(char **argv)

{
  printf("usage: %s [options]\n"
	 "where options might be:\n"
	 "\t-overhead-time      measure overhead for PCL (in time) \n"
         "\t-overhead-values    measure overhead for PCL (in counter values) \n"
	 "\t-startstop          do PCLstart/PCLstop\n"
	 "\t-read               do PCLstart/PCLread/PCLstop\n"
	 "\t-nested             do PCLstart/PCLstart/PCLstop/PCLstop\n"
	 "\t-multiple           do PCLstart(2 events)/PCLstop(2 events)\n"
	 "\t-overflow           generate 32-bit overflow\n"
	 "\t-mode (1|2|3)       user | system | user+system mode\n"
	 "\t-all                all of the above\n",
	 argv[0]);
  exit(1);
}


/*                                                                            */
/*============================================================================*/
/* main program                                                               */


int
main(int argc, char **argv)
{
  int i, j, n = N;
  int do_overhead_time, do_overhead_values, do_startstop, do_read, do_nested,
    do_multiple, do_overflow;
  int res, event, count_chosen, counter_list[2];
  unsigned int mode;
  double mhz;
  double a[N][N], b[N][N], c[N][N];
  struct timeval tv1, tv2;
  PCL_CNT_TYPE i_result[2] = {0, 0};
  PCL_FP_CNT_TYPE fp_result[2] = {0.0, 0.0};


  /* set default values */
#if defined(MIPS)
  mode = PCL_MODE_USER;
#else
  mode = PCL_MODE_USER_SYSTEM;
#endif
  do_startstop =
	  do_overhead_time =
	  do_overhead_values =
	  do_read =
	  do_nested =
	  do_multiple =
	  do_overflow = 0;
  count_chosen = 0;

  /* get MHz rate of processor */
  if((i = PCL_determine_mhz_rate()) == -1)
    {
      printf("not able to determine MHz rate of processor\n");
      mhz = 1.0;
    }
  else
  {
    mhz = (double)i;
    printf("processor running at %6.0f MHz\n", mhz);
  }

  /* parse arguments */
  if (argc > 1)
    {

      for (i = 1; i < argc; ++i)
	{
	  /* parse options */
	  if (!strcmp(argv[i], "-overhead-time"))
	    {
	      count_chosen = 1;
	      do_overhead_time = 1;
	    }

	  else if (!strcmp(argv[i], "-overhead-values"))
	    {
	      count_chosen = 1;
	      do_overhead_values = 1;
	    }

	  else if (!strcmp(argv[i], "-startstop"))
	    {
	      count_chosen = 1;
	      do_startstop = 1;
	    }

	  else if (!strcmp(argv[i], "-read"))
	    {
	      count_chosen = 1;
	      do_read = 1;
	    }

	  else if (!strcmp(argv[i], "-nested"))
	    {
	      count_chosen = 1;
	      do_nested = 1;
	    }

	  else if (!strcmp(argv[i], "-multiple"))
	    {
	      count_chosen = 1;
	      do_multiple = 1;
	    }

	  else if (!strcmp(argv[i], "-overflow"))
	    {
	      count_chosen = 1;
	      do_overflow = 1;
	    }

	  else if (!strcmp(argv[i], "-mode"))
	    {
	      ++i;
	      if ((i >= argc) || (sscanf(argv[i], "%d", &mode) != 1))
		usage(argv);
	    }

	  else if (!strcmp(argv[i], "-all"))
	    do_overhead_time =
		    do_overhead_values =
		    do_startstop =
		    do_read =
		    do_nested =
		    do_multiple =
		    do_overflow = 1;

	  else
	    usage(argv);
	}
    }


  if (count_chosen == 0)
    /* no counting chosen: choose default counting */
    do_startstop = 1;


  /* initialize test arrays */
  for (i = 0; i < n; ++i)
    for (j = 0; j < n; ++j)
      a[i][j] = b[i][j] = c[i][j] = 1.0;


  /* do some initial work */
  matadd(a, b, c);


/*-----------------------------------------------------------*/
  /* overhead (in time) */


  if (do_overhead_time)
    {
      /* define events to be measured */
      event = PCL_INSTR;
      counter_list[0] = event;

      gettimeofday(&tv1, NULL);
      for (i = 0; i < NITER * 100; ++i)
	dummy(event);
      gettimeofday(&tv2, NULL);

      printf("time for empty funcall       :%9.3f us/%9.0f cycles\n",
	     timediff(tv2, tv1) / (NITER * 100.0) * 1e6,
	     (timediff(tv2, tv1) / (NITER * 100.0) * 1e6) * mhz);


      /* query for functionality */
      if ((res = PCLquery(counter_list, 1, mode)) == PCL_SUCCESS)
	{
	  /* measure 1st level read */
	  /* start counting level */
	  if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
	    {
	      printf("problem with starting event %d (%s)\n",
		     event, errornames[-res]);
	      exit(1);
	    }

	  gettimeofday(&tv1, NULL);
	  for (i = 0; i < NITER; ++i)
	    PCLread(&i_result[0], &fp_result[0], 1);
	  gettimeofday(&tv2, NULL);
	  printf("time for PCLread             :%9.3f us/%9.0f cycles\n",
		 timediff(tv2, tv1) / NITER * 1e6,
		 (timediff(tv2, tv1) / NITER * 1e6) * mhz);


	  /* measure 1st level start/stop */
	  gettimeofday(&tv1, NULL);
	  for (i = 0; i < NITER; ++i)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* stop counting */
	      if ((res = PCLstop(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	    }
	  gettimeofday(&tv2, NULL);

	  printf("time for PCLstart/stop       :%9.3f us/%9.0f cycles\n",
		 timediff(tv2, tv1) / NITER * 1e6,
		 (timediff(tv2, tv1) / NITER * 1e6) * mhz);


	  /* measure 2nd level start/stop */
	  /* start counting level 1 */
	  if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
	    {
	      printf("problem with starting event %d (%s)\n",
		     event, errornames[-res]);
	      exit(1);
	    }
	  gettimeofday(&tv1, NULL);
	  for (i = 0; i < NITER; ++i)
	    {
	      /* start counting level 2 */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* stop counting level 2 */
	      if ((res = PCLstop(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	    }
	  gettimeofday(&tv2, NULL);

	  /* stop counting level 1 */
	  if ((res = PCLstop(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
	    printf("problem with stopping event %d (%s)\n",
		   event, errornames[-res]);

	  printf("time for nested PCLstart/stop:%9.3f us/%9.0f cycles\n",
		 timediff(tv2, tv1) / NITER * 1e6,
		 (timediff(tv2, tv1) / NITER * 1e6) * mhz);
	}

      else
	printf("%s\n", errornames[-res]);
    }


/*-----------------------------------------------------------*/
  /* overhead (with performance values) */


  if (do_overhead_values)
    {
      printf("overhead for inner start/stop\n");
      printf("%-26s %13s %13s\n", "overhead", "start/stop", "read");
      printf("---------\n");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-25s: ", PCLeventname(event));
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, mode)) == PCL_SUCCESS)
	    {
/*-------------------------------------------------------------*/
	      /* measure start/stop */

	      /* start counting level 0 */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* start counting level 1 */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* stop counting level 1 */
	      if ((res = PCLstop(&i_result[0], &fp_result[0], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);


	      /* stop counting level 0 */
	      if ((res = PCLstop(&i_result[0], &fp_result[0], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);


/*-------------------------------------------------------------*/
	      /* measure read */

	      /* start counting level */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* read values */
	      if ((res = PCLread(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
		printf("problem with reading event %d (%s)\n",
		       event, errornames[-res]);


	      /* stop counting level */
	      if ((res = PCLstop(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);

	      print_result(event, i_result[0], fp_result[0]);
	      printf(" ");
	      print_result(event, i_result[1], fp_result[1]);
	      printf("\n");
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test start/stop for all possible events */

  if (do_startstop)
    {
      printf("\ntesting start/stop for available events:\n");
      printf("----------------------------------------\n");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-25s: ", PCLeventname(event));
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, mode)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* stop counting */
	      if ((res = PCLstop(i_result, fp_result, 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		{
		  print_result(event, i_result[0], fp_result[0]);
		  printf("\n");
		}
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test read for all possible events */

  if (do_read)
    {
      printf("\ntesting read for available events:\n");
      printf("----------------------------------\n");
      printf("%-26s %13s %13s\n", "", "inner read", "outer stop");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-25s: ", PCLeventname(event));
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, mode)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* read values */
	      if ((res = PCLread(&i_result[0], &fp_result[0], 1)) != PCL_SUCCESS)
		printf("problem with reading event %d (%s)\n",
		       event, errornames[-res]);

	      /* do work */
	      matadd(a, b, c);

	      /* stop counting */
	      if ((res = PCLstop(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		{
		  print_result(event, i_result[0], fp_result[0]);
		  printf(" ");
		  print_result(event, i_result[1], fp_result[1]);
		  printf("\n");
		}
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test nested calls for all possible events */

  if (do_nested)
    {
      printf("\ntesting nested calls for available events:\n");
      printf("--------------------------------------------\n");
      printf("%-26s %13s %13s\n", "", "inner", "outer");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-25s: ", PCLeventname(event));
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, mode)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, mode)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* stop counting */
	      if ((res = PCLstop(&i_result[0], &fp_result[0], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);

	      /* stop counting */
	      if ((res = PCLstop(&i_result[1], &fp_result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		{
		  print_result(event, i_result[0], fp_result[0]);
		  printf(" ");
		  print_result(event, i_result[1], fp_result[1]);
		  printf("\n");
		}
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test for multiple events */

  if (do_multiple)
    {
      printf("\ntesting for multiple events:\n");
      printf("----------------------------\n");

      /* define events to be measured */
      counter_list[0] = PCL_CYCLES;
#if defined(SPARC)
      counter_list[1] = PCL_INSTR;
#else
      counter_list[1] = PCL_FP_INSTR;
#endif

      /* query for functionality */
      if ((res = PCLquery(counter_list, 2, mode)) != PCL_SUCCESS)
	printf("two events not possible (%s)\n", errornames[-res]);
      else
	{
	  /* start counting */
	  if ((res = PCLstart(counter_list, 2, mode)) != PCL_SUCCESS)
	    {
	      printf("problem with starting two events (%s)\n",
		     errornames[-res]);
	      exit(1);
	    }

	  /* do work */
	  matadd(a, b, c);

	  /* stop counting */
	  if ((res = PCLstop(i_result, fp_result, 2)) != PCL_SUCCESS)
	    printf("problem with stopping two events (%s)\n", errornames[-res]);
	  else
	    {
	      printf(OFORMAT " cycles with " OFORMAT
#if defined(SPARC)
	      " instructions\n",
#else
	      " floating point operations\n",
#endif
		     (OTYPE) i_result[0], (OTYPE) i_result[1]);
	    }
	}
    }

/*-----------------------------------------------------------*/
  /* test 32-bit overflow */

  if (do_overflow)
    {
#if defined(MIPS)
      double dijk = 0.0;
#endif

      printf("\ntesting 32-bit overflow (may take a while):\n");
      printf("-------------------------------------------\n");

      /* count cycles */
      counter_list[0] = PCL_CYCLES;
#if defined(MIPS)
      counter_list[1] = PCL_FP_INSTR;
#else
      counter_list[1] = PCL_ELAPSED_CYCLES;
#endif

      /* start counting */
      if ((res = PCLstart(counter_list, 2, mode)) != PCL_SUCCESS)
	{
	  printf("problem with starting event(%s)\n", errornames[-res]);
	  exit(1);
	}

      /* do work */
      {
	unsigned int i, j, k, ijk;

	/* We try to get 2^32 cycles for sure. */
	ijk = 0;
	for (i = 2048; i > 0; --i)
	  for (j = 2048; j > 0; --j)
	    for (k = 1024; k > 0; --k)
#if defined(MIPS)
	      dijk += i - j + k;
	ijk = dijk / 300;
#else
	      ijk += i - j + k;
#endif

	/* some compilers are smart */
	dummy(ijk);
      }

      /* stop counting */
      if ((res = PCLstop(i_result, fp_result, 2)) != PCL_SUCCESS)
	printf("problem with stopping event (%s)\n", errornames[-res]);
      else
#if defined(MIPS)
	printf(OFORMAT " cycles executed," OFORMAT " FP instr\n",
	       (OTYPE) i_result[0], (OTYPE) i_result[1]);
#else
	printf(OFORMAT " cycles executed," OFORMAT " elapsed cycles\n",
	       (OTYPE) i_result[0], (OTYPE) i_result[1]);
#endif
    }

/*-----------------------------------------------------------*/

  return 0;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */

void
dummy(unsigned int ijk)
{
}


/*                                                                            */
/*============================================================================*/
